# Derived from <https://gist.github.com/jart/0a71cde3ca7261f77080a3625a21672b>

# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import logging
import os.path
from pathlib import Path
import threading
from typing import Any, Callable, Iterable, Optional, Union

log = logging.getLogger(__name__)


def threaded_walk(
    dirpath: Union[str, Path],
    func: Optional[Callable[[Path], Any]] = None,
    threads: int = 60,
) -> Iterable[Any]:
    if not os.path.isdir(dirpath):
        return
    lock = threading.Lock()
    on_input = threading.Condition(lock)
    on_output = threading.Condition(lock)
    tasks = 1
    paths = [Path(dirpath)]
    output: list = []

    def worker() -> None:
        nonlocal tasks
        while True:
            with lock:
                while True:
                    if not tasks:
                        output.append(None)
                        on_output.notify()
                        return
                    if not paths:
                        on_input.wait()
                        continue
                    path = paths.pop()
                    break
            try:
                for p in path.iterdir():
                    if p.is_dir():
                        with lock:
                            tasks += 1
                            paths.append(p)
                            on_input.notify()
                    else:
                        item = func(p) if func is not None else p
                        with lock:
                            output.append(item)
                            on_output.notify()
            except Exception:
                log.exception("Error scanning directory %s", path)
            finally:
                with lock:
                    tasks -= 1
                    if not tasks:
                        on_input.notify_all()

    workers = [
        threading.Thread(
            target=worker, name=f"threaded_walk {i} {dirpath}", daemon=True
        )
        for i in range(threads)
    ]
    for w in workers:
        w.start()
    while threads or output:
        with lock:
            while not output:
                on_output.wait()
            item = output.pop()
        if item:
            yield item
        else:
            threads -= 1



from abc import ABC, abstractmethod
from collections import deque
from dataclasses import dataclass, field, replace
from datetime import datetime
from enum import Enum
import json
import os.path
from pathlib import Path, PurePosixPath
import re
from time import sleep, time
from types import TracebackType
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)
from urllib.parse import unquote, urlparse, urlunparse

import click
from dandischema import models
import dateutil.parser
from pydantic import AnyHttpUrl, BaseModel, Field, PrivateAttr
import requests
import tenacity

from . import get_logger
from .consts import (
    DRAFT,
    MAX_CHUNK_SIZE,
    RETRY_STATUSES,
    ZARR_DELETE_BATCH_SIZE,
    DandiInstance,
    EmbargoStatus,
    known_instances,
    known_instances_rev,
)
from .exceptions import HTTP404Error, NotFoundError, SchemaVersionError
from .keyring import keyring_lookup
from .misctypes import BasePath, Digest
from .utils import (
    USER_AGENT,
    check_dandi_version,
    chunked,
    ensure_datetime,
    is_interactive,
)

lgr = get_logger()

T = TypeVar("T")


class AssetType(Enum):
    """
    .. versionadded:: 0.36.0

    An enum for the different kinds of resources that an asset's actual data
    can be
    """

    BLOB = 1
    ZARR = 2


# Following class is loosely based on GirderClient, with authentication etc
# being stripped.
# TODO: add copyright/license info
class RESTFullAPIClient:
    """
    Base class for a JSON-based HTTP(S) client for interacting with a given
    base API URL.

    All request methods can take either an absolute URL or a slash-separated
    path; in the latter case, the path is appended to the base API URL
    (separated by a slash) in order to determine the actual URL to make the
    request of.

    `RESTFullAPIClient` instances are usable as context managers, in which case
    they will close their associated session on exit.
    """

    def __init__(
        self,
        api_url: str,
        session: Optional[requests.Session] = None,
        headers: Optional[dict] = None,
    ) -> None:
        """
        :param str api_url: The base HTTP(S) URL to prepend to request paths
        :param session: an optional `requests.Session` instance to use; if not
            specified, a new session is created
        :param headers: an optional `dict` of headers to send in every request
        """
        self.api_url = api_url
        if session is None:
            session = requests.Session()
        session.headers["User-Agent"] = USER_AGENT
        if headers is not None:
            session.headers.update(headers)
        self.session = session

    def __enter__(self: T) -> T:
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.session.close()

    def request(
        self,
        method: str,
        path: str,
        params: Optional[dict] = None,
        data: Any = None,
        files: Optional[dict] = None,
        json: Any = None,
        headers: Optional[dict] = None,
        json_resp: bool = True,
        retry_statuses: Sequence[int] = (),
        **kwargs: Any,
    ) -> Any:
        """
        This method looks up the appropriate method, constructs a request URL
        from the base URL, path, and parameters, and then sends the request. If
        the method is unknown or if the path is not found, an exception is
        raised; otherwise, a JSON object is returned with the response.

        This is a convenience method to use when making basic requests that do
        not involve multipart file data that might need to be specially encoded
        or handled differently.

        :param method: The HTTP method to use in the request (GET, POST, etc.)
        :type method: str
        :param path: A string containing the path elements for this request
        :type path: str
        :param params: A dictionary mapping strings to strings, to be used
            as the key/value pairs in the request parameters.
        :type params: dict
        :param data: A dictionary, bytes or file-like object to send in the
            body.
        :param files: A dictionary of 'name' => file-like-objects for multipart
            encoding upload.
        :type files: dict
        :param json: A JSON object to send in the request body.
        :type json: dict
        :param headers: If present, a dictionary of headers to encode in the
            request.
        :type headers: dict
        :param json_resp: Whether the response should be parsed as JSON. If
            False, the raw response object is returned. To get the raw binary
            content of the response, use the ``content`` attribute of the
            return value, e.g.

            .. code-block:: python

                resp = client.get('my/endpoint', json_resp=False)
                print(resp.content)  # Raw binary content
                print(resp.headers)  # Dict of headers

        :type json_resp: bool
        :param retry_statuses: a sequence of HTTP response status codes to
            retry in addition to `dandi.consts.RETRY_STATUSES`
        """

        url = self.get_url(path)

        if headers is None:
            headers = {}
        if json_resp and "accept" not in headers:
            headers["accept"] = "application/json"

        lgr.debug("%s %s", method.upper(), url)

        try:
            for i, attempt in enumerate(
                tenacity.Retrying(
                    wait=tenacity.wait_exponential(exp_base=1.25, multiplier=1.25),
                    # urllib3's ConnectionPool isn't thread-safe, so we
                    # sometimes hit ConnectionErrors on the start of an upload.
                    # Retry when this happens.
                    # Cf. <https://github.com/urllib3/urllib3/issues/951>.
                    retry=tenacity.retry_if_exception_type(
                        (requests.ConnectionError, requests.HTTPError)
                    ),
                    stop=tenacity.stop_after_attempt(12),
                    reraise=True,
                )
            ):
                with attempt:
                    if attempt.retry_state.attempt_number > 1:
                        lgr.warning("Retrying %s %s", method.upper(), url)
                    result = self.session.request(
                        method,
                        url,
                        params=params,
                        data=data,
                        files=files,
                        json=json,
                        headers=headers,
                        **kwargs,
                    )
                    if result.status_code in [*RETRY_STATUSES, *retry_statuses]:
                        result.raise_for_status()
        except Exception:
            lgr.exception("HTTP connection failed")
            raise

        if i > 0:
            lgr.info(
                "%s %s succeeded after %d retr%s",
                method.upper(),
                url,
                i,
                "y" if i == 1 else "ies",
            )

        lgr.debug("Response: %d", result.status_code)

        # If success, return the json object. Otherwise throw an exception.
        if not result.ok:
            msg = f"Error {result.status_code} while sending {method} request to {url}"
            if result.status_code == 409:
                # Blob exists on server; log at DEBUG level
                lgr.debug("%s: %s", msg, result.text)
            else:
                lgr.error("%s: %s", msg, result.text)
            if result.status_code == 404:
                raise HTTP404Error(msg, response=result)
            else:
                raise requests.HTTPError(msg, response=result)

        if json_resp:
            if result.text.strip():
                return result.json()
            else:
                return None
        else:
            return result

    def get_url(self, path: str) -> str:
        """
        Append a slash-separated ``path`` to the instance's base URL.  The two
        components are separated by a single slash, and any trailing slashes
        are removed.

        If ``path`` is already an absolute URL, it is returned unchanged.
        """
        # Construct the url
        if path.lower().startswith(("http://", "https://")):
            return path
        else:
            return self.api_url.rstrip("/") + "/" + path.lstrip("/")

    def get(self, path: str, **kwargs: Any) -> Any:
        """
        Convenience method to call `request()` with the 'GET' HTTP method.
        """
        return self.request("GET", path, **kwargs)

    def post(self, path: str, **kwargs: Any) -> Any:
        """
        Convenience method to call `request()` with the 'POST' HTTP method.
        """
        return self.request("POST", path, **kwargs)

    def put(self, path: str, **kwargs: Any) -> Any:
        """
        Convenience method to call `request()` with the 'PUT' HTTP method.
        """
        return self.request("PUT", path, **kwargs)

    def delete(self, path: str, **kwargs: Any) -> Any:
        """
        Convenience method to call `request()` with the 'DELETE' HTTP method.
        """
        return self.request("DELETE", path, **kwargs)

    def patch(self, path: str, **kwargs: Any) -> Any:
        """
        Convenience method to call `request()` with the 'PATCH' HTTP method.
        """
        return self.request("PATCH", path, **kwargs)

    def paginate(
        self,
        path: str,
        page_size: Optional[int] = None,
        params: Optional[dict] = None,
        **kwargs: Any,
    ) -> Iterator:
        """
        Paginate through the resources at the given path: GET the path, yield
        the values in the ``"results"`` key, and repeat with the URL in the
        ``"next"`` key until it is ``null``.
        """
        if page_size is not None:
            if params is None:
                params = {}
            params["page_size"] = page_size
        r = self.get(path, params=params, **kwargs)
        while True:
            for item in r["results"]:
                yield item
            if r.get("next"):
                r = self.get(r["next"])
            else:
                break


class DandiAPIClient(RESTFullAPIClient):
    """A client for interacting with a Dandi Archive server"""

    def __init__(
        self, api_url: Optional[str] = None, token: Optional[str] = None
    ) -> None:
        """
        Construct a client instance.

        :param str api_url: Base API URL of the server to interact with.
        - For DANDI production, use  ``"https://api.dandiarchive.org/api"``
        - For DANDI staging, use ``"https://api-staging.dandiarchive.org/api"``
        - If no URL is supplied, the URL is looked up in `known_instances` using the value of the
        :envvar:`DANDI_INSTANCE` environment variable (default value: ``"dandi"``).
        :param str token: User API Key. Note that different instance APIs have different
        keys.
        """

        check_dandi_version()
        if api_url is None:
            instance_name = os.environ.get("DANDI_INSTANCE", "dandi")
            api_url = known_instances[instance_name].api
            if api_url is None:
                raise ValueError(f"No API URL for instance {instance_name!r}")
        super().__init__(api_url)
        if token is not None:
            self.authenticate(token)

    @classmethod
    def for_dandi_instance(
        cls,
        instance: Union[str, DandiInstance],
        token: Optional[str] = None,
        authenticate: bool = False,
    ) -> "DandiAPIClient":
        """
        Construct a client instance for the server identified by ``instance``
        (either the name of a registered Dandi Archive instance or a
        `DandiInstance` instance) and an optional authentication token/API key.
        If no token is supplied and ``authenticate`` is true,
        `dandi_authenticate()` is called on the instance before returning it.
        """
        if isinstance(instance, str):
            instance = known_instances[instance]
        client = cls(instance.api, token=token)
        if token is None and authenticate:
            client.dandi_authenticate()
        return client

    def authenticate(self, token: str) -> None:
        """
        Set the authentication token/API key used by the `DandiAPIClient`.
        Before setting the token, a test request to ``/auth/token`` is made to
        check the token's validity; if it fails, a `requests.HTTPError` is
        raised.
        """
        # Fails if token is invalid:
        self.get("/auth/token", headers={"Authorization": f"token {token}"})
        self.session.headers["Authorization"] = f"token {token}"

    def dandi_authenticate(self) -> None:
        """
        Acquire and set the authentication token/API key used by the
        `DandiAPIClient`.  If the :envvar:`DANDI_API_KEY` environment variable
        is set, its value is used as the token.  Otherwise, the token is looked
        up in the user's keyring under the service
        ":samp:`dandi-api-{INSTANCE_NAME}`" [#auth]_ and username "``key``".
        If no token is found there, the user is prompted for the token, and, if
        it proves to be valid, it is stored in the user's keyring.

        .. [#auth] E.g., "``dandi-api-dandi``" for the production server or
                   "``dandi-api-dandi-staging``" for the staging server
        """
        # Shortcut for advanced folks
        api_key = os.environ.get("DANDI_API_KEY", None)
        if api_key:
            self.authenticate(api_key)
            return
        if self.api_url in known_instances_rev:
            client_name = known_instances_rev[self.api_url]
        else:
            raise NotImplementedError("TODO client name derivation for keyring")
        app_id = f"dandi-api-{client_name}"
        keyring_backend, api_key = keyring_lookup(app_id, "key")
        key_from_keyring = api_key is not None
        while True:
            if not api_key:
                api_key = input(f"Please provide API Key for {client_name}: ")
                key_from_keyring = False
            try:
                self.authenticate(api_key)
            except requests.HTTPError:
                if is_interactive() and click.confirm(
                    "API key is invalid; enter another?"
                ):
                    api_key = None
                    continue
                else:
                    raise
            else:
                if not key_from_keyring:
                    keyring_backend.set_password(app_id, "key", api_key)
                    lgr.debug("Stored key in keyring")
                break

    @property
    def _instance_id(self) -> str:
        url = self.api_url.rstrip("/")
        try:
            return known_instances_rev[url].upper()
        except KeyError:
            return url

    def get_dandiset(
        self, dandiset_id: str, version_id: Optional[str] = None, lazy: bool = True
    ) -> "RemoteDandiset":
        """
        Fetches the Dandiset with the given ``dandiset_id``.  If ``version_id``
        is not specified, the `RemoteDandiset`'s version is set to the most
        recent published version if there is one, otherwise to the draft
        version.

        If ``lazy`` is true, no requests are actually made until any data is
        requested from the `RemoteDandiset`.
        """
        if lazy:
            return RemoteDandiset(self, dandiset_id, version_id)
        else:
            try:
                d = RemoteDandiset.from_data(
                    self, self.get(f"/dandisets/{dandiset_id}/")
                )
            except HTTP404Error:
                raise NotFoundError(f"No such Dandiset: {dandiset_id!r}")
            if version_id is not None and version_id != d.version_id:
                if version_id == DRAFT:
                    return d.for_version(d.draft_version)
                else:
                    return d.for_version(version_id)
            return d

    def get_dandisets(self) -> Iterator["RemoteDandiset"]:
        """
        Returns a generator of all Dandisets on the server.  For each Dandiset,
        the `RemoteDandiset`'s version is set to the most recent published
        version if there is one, otherwise to the draft version.
        """
        for data in self.paginate("/dandisets/"):
            yield RemoteDandiset.from_data(self, data)

    def create_dandiset(self, name: str, metadata: Dict[str, Any]) -> "RemoteDandiset":
        """Creates a Dandiset with the given name & metadata"""
        return RemoteDandiset.from_data(
            self, self.post("/dandisets/", json={"name": name, "metadata": metadata})
        )

    def check_schema_version(self, schema_version: Optional[str] = None) -> None:
        """
        Confirms that the server is using the same version of the Dandi schema
        as the client.  If it is not, a `SchemaVersionError` is raised.

        :param schema_version: the schema version to confirm that the server
            uses; if not set, the schema version for the installed
            ``dandischema`` library is used
        """
        if schema_version is None:
            schema_version = models.get_schema_version()
        server_info = self.get("/info/")
        server_schema_version = server_info.get("schema_version")
        if not server_schema_version:
            raise RuntimeError(
                "Server did not provide schema_version in /info/;"
                f" returned {server_info!r}"
            )
        if server_schema_version != schema_version:
            raise SchemaVersionError(
                f"Server requires schema version {server_schema_version};"
                f" client only supports {schema_version}.  You may need to"
                " upgrade dandi and/or dandischema."
            )

    def get_asset(self, asset_id: str) -> "BaseRemoteAsset":
        """
        Fetch the asset with the given asset ID.  If the given asset does not
        exist, a `NotFoundError` is raised.

        The returned object will not have any information about the Dandiset
        associated with the asset; for that, the `RemoteDandiset.get_asset()`
        method must be used instead.
        """
        try:
            return BaseRemoteAsset.from_base_data(
                self, self.get(f"/assets/{asset_id}/info/")
            )
        except HTTP404Error:
            raise NotFoundError(f"No such asset: {asset_id!r}")


class APIBase(BaseModel):
    """
    Base class for API objects implemented in pydantic.

    This class (aside from the `json_dict()` method) is an implementation
    detail; do not rely on it.
    """

    JSON_EXCLUDE: ClassVar[FrozenSet[str]] = frozenset(["client"])

    def json_dict(self) -> Dict[str, Any]:
        """
        Convert to a JSONable `dict`, omitting the ``client`` attribute and
        using the same field names as in the API
        """
        return cast(
            Dict[str, Any],
            json.loads(self.json(exclude=self.JSON_EXCLUDE, by_alias=True)),
        )

    class Config:
        allow_population_by_field_name = True
        # To allow `client: Session`:
        arbitrary_types_allowed = True


class Version(APIBase):
    """
    The version information for a Dandiset retrieved from the API.

    Stringifying a `Version` returns its identifier.

    This class should not be instantiated by end-users directly.  Instead,
    instances should be retrieved from the appropriate attributes & methods of
    `RemoteDandiset`.
    """

    #: The version identifier
    identifier: str = Field(alias="version")
    #: The name of the version
    name: str
    #: The number of assets in the version
    asset_count: int
    #: The total size in bytes of all assets in the version
    size: int
    #: The timestamp at which the version was created
    created: datetime
    #: The timestamp at which the version was last modified
    modified: datetime

    def __str__(self) -> str:
        return self.identifier


class RemoteDandisetData(APIBase):
    """
    Class for storing the data for a Dandiset retrieved from the API.

    This class is an implementation detail and should not be used by third
    parties.
    """

    identifier: str
    created: datetime
    modified: datetime
    contact_person: str
    embargo_status: EmbargoStatus
    most_recent_published_version: Optional[Version]
    draft_version: Version


class RemoteDandiset:
    """
    Representation of a Dandiset (as of a certain version) retrieved from the
    API.

    Stringifying a `RemoteDandiset` returns a string of the form
    :samp:`"{server_id}:{dandiset_id}/{version_id}"`.

    This class should not be instantiated by end-users directly.  Instead,
    instances should be retrieved from the appropriate attributes & methods of
    `DandiAPIClient` and `RemoteDandiset`.
    """

    def __init__(
        self,
        client: DandiAPIClient,
        identifier: str,
        version: Union[str, Version, None] = None,
        data: Union[Dict[str, Any], RemoteDandisetData, None] = None,
    ) -> None:
        #: The `DandiAPIClient` instance that returned this `RemoteDandiset`
        #: and which the latter will use for API requests
        self.client: DandiAPIClient = client
        #: The Dandiset identifier
        self.identifier: str = identifier
        self._version_id: Optional[str]
        self._version: Optional[Version]
        if version is None:
            self._version_id = None
            self._version = None
        elif isinstance(version, str):
            self._version_id = version
            self._version = None
        else:
            self._version_id = version.identifier
            self._version = version
        self._data: Optional[RemoteDandisetData]
        if data is not None:
            self._data = RemoteDandisetData.parse_obj(data)
        else:
            self._data = None

    def __str__(self) -> str:
        return f"{self.client._instance_id}:{self.identifier}/{self.version_id}"

    def _get_data(self) -> RemoteDandisetData:
        if self._data is None:
            try:
                self._data = RemoteDandisetData.parse_obj(
                    self.client.get(f"/dandisets/{self.identifier}/")
                )
            except HTTP404Error:
                raise NotFoundError(f"No such Dandiset: {self.identifier}")
        return self._data

    @property
    def version_id(self) -> str:
        """The identifier for the Dandiset version"""
        if self._version_id is None:
            self._version_id = self.version.identifier
        return self._version_id

    @property
    def version(self) -> Version:
        """The version in question of the Dandiset"""
        if self._version is None:
            if self._version_id is None:
                self._get_data()
            if self._data is not None:
                for v in [
                    self._data.most_recent_published_version,
                    self._data.draft_version,
                ]:
                    if v is not None and (
                        self._version_id is None or v.identifier == self.version_id
                    ):
                        self._version = v
                        self._version_id = v.identifier
                        return v
            assert self._version_id is not None
            self._version = self.get_version(self._version_id)
        return self._version

    @property
    def created(self) -> datetime:
        """The timestamp at which the Dandiset was created"""
        return self._get_data().created

    @property
    def modified(self) -> datetime:
        """The timestamp at which the Dandiset was last modified"""
        return self._get_data().modified

    @property
    def contact_person(self) -> str:
        """The name of the registered contact person for the Dandiset"""
        return self._get_data().contact_person

    @property
    def embargo_status(self) -> EmbargoStatus:
        """The current embargo status for the Dandiset"""
        return self._get_data().embargo_status

    @property
    def most_recent_published_version(self) -> Optional[Version]:
        """
        The most recent published (non-draft) version of the Dandiset, or
        `None` if no versions have been published
        """
        return self._get_data().most_recent_published_version

    @property
    def draft_version(self) -> Version:
        """The draft version of the Dandiset"""
        return self._get_data().draft_version

    @property
    def api_path(self) -> str:
        """
        The path (relative to the base endpoint for a Dandi Archive API) at
        which API requests for interacting with the Dandiset itself are made
        """
        return f"/dandisets/{self.identifier}/"

    @property
    def api_url(self) -> str:
        """
        The URL at which API requests for interacting with the Dandiset itself
        are made
        """
        return self.client.get_url(self.api_path)

    @property
    def version_api_path(self) -> str:
        """
        The path (relative to the base endpoint for a Dandi Archive API) at
        which API requests for interacting with the version in question of the
        Dandiset are made
        """
        return f"/dandisets/{self.identifier}/versions/{self.version_id}/"

    @property
    def version_api_url(self) -> str:
        """
        The URL at which API requests for interacting with the version in
        question of the Dandiset are made
        """
        return self.client.get_url(self.version_api_path)

    @classmethod
    def from_data(
        cls, client: "DandiAPIClient", data: Dict[str, Any]
    ) -> "RemoteDandiset":
        """
        Construct a `RemoteDandiset` instance from a `DandiAPIClient` and a
        `dict` of raw string fields in the same format as returned by the API.
        If the ``"most_recent_published_version"`` field is set, that is used
        as the Dandiset's version; otherwise, ``"draft_version"`` is used.

        This is a low-level method that non-developers would normally only use
        when acquiring data using means outside of this library.
        """
        if data.get("most_recent_published_version") is not None:
            version = Version.parse_obj(data["most_recent_published_version"])
        else:
            version = Version.parse_obj(data["draft_version"])
        return cls(
            client=client, identifier=data["identifier"], version=version, data=data
        )

    def json_dict(self) -> Dict[str, Any]:
        """
        Convert to a JSONable `dict`, omitting the ``client`` attribute and
        using the same field names as in the API
        """
        return {
            **self._get_data().json_dict(),
            "version": self.version.json_dict(),
        }

    def refresh(self) -> None:
        """
        Update the `RemoteDandiset` in-place with the latest data from the
        server.  The `RemoteDandiset` continues to have the same version as
        before, but the cached version data is internally cleared and may be
        different upon subsequent access.
        """
        self._data = None
        self._get_data()
        # Clear _version so it will be refetched the next time it is accessed
        self._version = None

    def get_versions(self) -> Iterator[Version]:
        """Returns an iterator of all available `Version`\\s for the Dandiset"""
        try:
            for v in self.client.paginate(f"{self.api_path}versions/"):
                yield Version.parse_obj(v)
        except HTTP404Error:
            raise NotFoundError(f"No such Dandiset: {self.identifier!r}")

    def get_version(self, version_id: str) -> Version:
        """
        Get information about a given version of the Dandiset.  If the given
        version does not exist, a `NotFoundError` is raised.
        """
        try:
            return Version.parse_obj(
                self.client.get(
                    f"/dandisets/{self.identifier}/versions/{version_id}/info"
                )
            )
        except HTTP404Error:
            raise NotFoundError(
                f"No such version: {version_id!r} of Dandiset {self.identifier}"
            )

    def for_version(self, version_id: Union[str, Version]) -> "RemoteDandiset":
        """
        Returns a copy of the `RemoteDandiset` with the `version` attribute set
        to the given `Version` object or the `Version` with the given version
        ID.  If a version ID given and the version does not exist, a
        `NotFoundError` is raised.
        """
        if isinstance(version_id, str):
            version_id = self.get_version(version_id)
        return type(self)(
            client=self.client,
            identifier=self.identifier,
            version=version_id,
            data=self._data,
        )

    def delete(self) -> None:
        """
        Delete the Dandiset from the server.  Any further access of the
        instance's data attributes afterwards will result in a `NotFoundError`.
        """
        self.client.delete(self.api_path)
        self._data = None
        self._version = None

    def get_metadata(self) -> models.Dandiset:
        """
        Fetch the metadata for this version of the Dandiset as a
        `dandischema.models.Dandiset` instance
        """
        return models.Dandiset.parse_obj(self.get_raw_metadata())

    def get_raw_metadata(self) -> Dict[str, Any]:
        """
        Fetch the metadata for this version of the Dandiset as an unprocessed
        `dict`
        """
        try:
            return cast(Dict[str, Any], self.client.get(self.version_api_path))
        except HTTP404Error:
            raise NotFoundError(f"No such asset: {self}")

    def set_metadata(self, metadata: models.Dandiset) -> None:
        """
        Set the metadata for this version of the Dandiset to the given value
        """
        self.set_raw_metadata(metadata.json_dict())

    def set_raw_metadata(self, metadata: Dict[str, Any]) -> None:
        """
        Set the metadata for this version of the Dandiset to the given value
        """
        self.client.put(
            self.version_api_path,
            json={"metadata": metadata, "name": metadata.get("name", "")},
        )

    def wait_until_valid(self, max_time: float = 120) -> None:
        """
        Wait at most ``max_time`` seconds for the Dandiset to be valid for
        publication.  If the Dandiset does not become valid in time, a
        `ValueError` is raised.
        """
        lgr.debug("Waiting for Dandiset %s to complete validation ...", self.identifier)
        start = time()
        while time() - start < max_time:
            try:
                r = self.client.get(f"{self.version_api_path}info/")
            except HTTP404Error:
                raise NotFoundError(
                    f"No such version: {self.version_id!r} of Dandiset {self.identifier}"
                )
            if "status" not in r:
                # Running against older version of dandi-api that doesn't
                # validate
                return
            if r["status"] == "Valid":
                return
            sleep(0.5)
        # TODO: Improve the presentation of the error messages
        about = {
            "asset_validation_errors": r.get("asset_validation_errors"),
            "version_validation_errors": r.get("version_validation_errors"),
        }
        raise ValueError(
            f"Dandiset {self.identifier} is {r['status']}: {json.dumps(about, indent=4)}"
        )

    def publish(self) -> "RemoteDandiset":
        """
        Publish this version of the Dandiset.  Returns a copy of the
        `RemoteDandiset` with the `version` attribute set to the new published
        `Version`.
        """
        return self.for_version(
            Version.parse_obj(self.client.post(f"{self.version_api_path}publish/"))
        )

    def get_assets(self, order: Optional[str] = None) -> Iterator["RemoteAsset"]:
        """
        Returns an iterator of all assets in this version of the Dandiset.

        Assets can be sorted by a given field by passing the name of that field
        as the ``order`` parameter.  The accepted field names are
        ``"created"``, ``"modified"``, and ``"path"``.  Prepend a hyphen to the
        field name to reverse the sort order.
        """
        try:
            for a in self.client.paginate(
                f"{self.version_api_path}assets/", params={"order": order}
            ):
                yield RemoteAsset.from_data(self, a)
        except HTTP404Error:
            raise NotFoundError(
                f"No such version: {self.version_id!r} of Dandiset {self.identifier}"
            )

    def get_asset(self, asset_id: str) -> "RemoteAsset":
        """
        Fetch the asset in this version of the Dandiset with the given asset
        ID.  If the given asset does not exist, a `NotFoundError` is raised.
        """
        try:
            metadata = self.client.get(f"{self.version_api_path}assets/{asset_id}/")
        except HTTP404Error:
            raise NotFoundError(f"No such asset: {asset_id!r} for {self}")
        asset = self.get_asset_by_path(metadata["path"])
        asset._metadata = metadata
        return asset

    def get_assets_with_path_prefix(
        self, path: str, order: Optional[str] = None
    ) -> Iterator["RemoteAsset"]:
        """
        Returns an iterator of all assets in this version of the Dandiset whose
        `~RemoteAsset.path` attributes start with ``path``

        Assets can be sorted by a given field by passing the name of that field
        as the ``order`` parameter.  The accepted field names are
        ``"created"``, ``"modified"``, and ``"path"``.  Prepend a hyphen to the
        field name to reverse the sort order.
        """
        try:
            for a in self.client.paginate(
                f"{self.version_api_path}assets/",
                params={"path": path, "order": order},
            ):
                yield RemoteAsset.from_data(self, a)
        except HTTP404Error:
            raise NotFoundError(
                f"No such version: {self.version_id!r} of Dandiset {self.identifier}"
            )

    def get_asset_by_path(self, path: str) -> "RemoteAsset":
        """
        Fetch the asset in this version of the Dandiset whose
        `~RemoteAsset.path` equals ``path``.  If the given asset does not
        exist, a `NotFoundError` is raised.
        """
        try:
            # Weed out any assets that happen to have the given path as a
            # proper prefix:
            (asset,) = (
                a for a in self.get_assets_with_path_prefix(path) if a.path == path
            )
        except ValueError:
            raise NotFoundError(f"No asset at path {path!r}")
        else:
            return asset

    def download_directory(
        self,
        assets_dirpath: str,
        dirpath: Union[str, Path],
        chunk_size: int = MAX_CHUNK_SIZE,
    ) -> None:
        """
        Download all assets under the virtual directory ``assets_dirpath`` to
        the directory ``dirpath``.  Downloads are synchronous.
        """
        if assets_dirpath and not assets_dirpath.endswith("/"):
            assets_dirpath += "/"
        assets = list(self.get_assets_with_path_prefix(assets_dirpath))
        for a in assets:
            filepath = Path(dirpath, a.path[len(assets_dirpath) :])
            filepath.parent.mkdir(parents=True, exist_ok=True)
            a.download(filepath, chunk_size=chunk_size)

    def upload_raw_asset(
        self,
        filepath: Union[str, Path],
        asset_metadata: Dict[str, Any],
        jobs: Optional[int] = None,
        replace_asset: Optional["RemoteAsset"] = None,
    ) -> "RemoteAsset":
        """
        Upload the file at ``filepath`` with metadata ``asset_metadata`` to
        this version of the Dandiset and return the resulting asset.  Blocks
        until the upload is complete.

        .. deprecated:: 0.36.0
            Use the ``upload()`` method of `~dandi.files.LocalAsset` instances
            instead

        :param filepath: the path to the local file to upload
        :type filepath: str or PathLike
        :param dict asset_metadata:
            Metadata for the uploaded asset file.  Must include a "path" field
            giving the forward-slash-separated path at which the uploaded file
            will be placed on the server.
        :param int jobs: Number of threads to use for uploading; defaults to 5
        :param RemoteAsset replace_asset: If set, replace the given asset,
            which must have the same path as the new asset
        """
        from .files import LocalAsset, dandi_file

        df = dandi_file(filepath)
        if not isinstance(df, LocalAsset):
            raise ValueError(f"{filepath}: not an asset file")
        return df.upload(
            self, metadata=asset_metadata, jobs=jobs, replacing=replace_asset
        )

    def iter_upload_raw_asset(
        self,
        filepath: Union[str, Path],
        asset_metadata: Dict[str, Any],
        jobs: Optional[int] = None,
        replace_asset: Optional["RemoteAsset"] = None,
    ) -> Iterator[dict]:
        """
        Upload the file at ``filepath`` with metadata ``asset_metadata`` to
        this version of the Dandiset, returning a generator of status
        `dict`\\s.

        .. deprecated:: 0.36.0
            Use the ``iter_upload()`` method of `~dandi.files.LocalAsset`
            instances instead

        :param filepath: the path to the local file to upload
        :type filepath: str or PathLike
        :param dict asset_metadata:
            Metadata for the uploaded asset file.  Must include a "path" field
            giving the forward-slash-separated path at which the uploaded file
            will be placed on the server.
        :param int jobs:
            Number of threads to use for uploading; defaults to 5
        :param RemoteAsset replace_asset: If set, replace the given asset,
            which must have the same path as the new asset
        :returns:
            A generator of `dict`\\s containing at least a ``"status"`` key.
            Upon successful upload, the last `dict` will have a status of
            ``"done"`` and an ``"asset"`` key containing the resulting
            `RemoteAsset`.
        """
        from .files import LocalAsset, dandi_file

        df = dandi_file(filepath)
        if not isinstance(df, LocalAsset):
            raise ValueError(f"{filepath}: not an asset file")
        return df.iter_upload(
            self, metadata=asset_metadata, jobs=jobs, replacing=replace_asset
        )


class BaseRemoteAsset(ABC, APIBase):
    """
    Representation of an asset retrieved from the API without associated
    Dandiset information.

    This is an abstract class; its concrete subclasses are
    `BaseRemoteBlobAsset` (for assets backed by blobs) and
    `BaseRemoteZarrAsset` (for assets backed by Zarrs).

    Stringifying a `BaseRemoteAsset` returns a string of the form
    :samp:`"{server_id}:asset/{asset_id}"`.

    This class should not be instantiated by end-users directly.  Instead,
    instances should be retrieved from the appropriate methods of
    `DandiAPIClient`.
    """

    #: The `DandiAPIClient` instance that returned this `BaseRemoteAsset`
    #: and which the latter will use for API requests
    client: "DandiAPIClient"
    #: The asset identifier
    identifier: str = Field(alias="asset_id")
    #: The asset's (forward-slash-separated) path
    path: str
    #: The size of the asset in bytes
    size: int
    #: The date at which the asset was created
    created: datetime
    #: The date at which the asset was last modified
    modified: datetime
    #: Metadata supplied at initialization; returned when metadata is requested
    #: instead of performing an API call
    _metadata: Optional[Dict[str, Any]] = PrivateAttr(default_factory=None)

    def __init__(self, **data: Any) -> None:  # type: ignore[no-redef]
        super().__init__(**data)
        # Pydantic insists on not initializing any attributes that start with
        # underscores, so we have to do it ourselves.
        self._metadata = data.get("metadata", data.get("_metadata"))

    def __str__(self) -> str:
        return f"{self.client._instance_id}:assets/{self.identifier}"

    @classmethod
    def from_base_data(
        self,
        client: "DandiAPIClient",
        data: Dict[str, Any],
        metadata: Optional[Dict[str, Any]] = None,
    ) -> "BaseRemoteAsset":
        """
        Construct a `BaseRemoteAsset` instance from a `DandiAPIClient`, a
        `dict` of raw data in the same format as returned by the API's
        pagination endpoints, and optional raw asset metadata.

        This is a low-level method that non-developers would normally only use
        when acquiring data using means outside of this library.
        """
        klass: Type[BaseRemoteAsset]
        if data.get("blob") is not None:
            klass = BaseRemoteBlobAsset
            if data.pop("zarr", None) is not None:
                raise ValueError("Asset data contains both `blob` and `zarr`'")
        elif data.get("zarr") is not None:
            klass = BaseRemoteZarrAsset
            if data.pop("blob", None) is not None:
                raise ValueError("Asset data contains both `blob` and `zarr`'")
        else:
            raise ValueError("Asset data contains neither `blob` nor `zarr`")
        return klass(client=client, **data, _metadata=metadata)  # type: ignore[call-arg]

    @property
    def api_path(self) -> str:
        """
        The path (relative to the base endpoint for a Dandi Archive API) at
        which API requests for interacting with the asset itself are made
        """
        return f"/assets/{self.identifier}/"

    @property
    def api_url(self) -> str:
        """
        The URL at which API requests for interacting with the asset itself are
        made
        """
        return self.client.get_url(self.api_path)

    @property
    def base_download_url(self) -> str:
        """
        The URL from which the asset can be downloaded, sans any Dandiset
        identifiers (cf. `RemoteAsset.download_url`)
        """
        return self.client.get_url(f"/assets/{self.identifier}/download/")

    def get_metadata(self) -> models.Asset:
        """
        Fetch the metadata for the asset as a `dandischema.models.Asset`
        instance
        """
        return models.Asset.parse_obj(self.get_raw_metadata())

    def get_raw_metadata(self) -> Dict[str, Any]:
        """Fetch the metadata for the asset as an unprocessed `dict`"""
        if self._metadata is not None:
            return self._metadata
        else:
            try:
                return cast(Dict[str, Any], self.client.get(self.api_path))
            except HTTP404Error:
                raise NotFoundError(f"No such asset: {self}")

    def get_raw_digest(
        self, digest_type: Union[str, models.DigestType, None] = None
    ) -> str:
        """
        Retrieves the value of the given type of digest from the asset's
        metadata.  Raises `NotFoundError` if there is no entry for the given
        digest type.

        If no digest type is specified, the same type as used by `get_digest()`
        is returned.

        .. versionchanged:: 0.36.0
            Renamed from ``get_digest()`` to ``get_raw_digest()``
        """
        if digest_type is None:
            digest_type = self.digest_type.value
        elif isinstance(digest_type, models.DigestType):
            digest_type = digest_type.value
        metadata = self.get_raw_metadata()
        try:
            return cast(str, metadata["digest"][digest_type])
        except KeyError:
            raise NotFoundError(f"No {digest_type} digest found in metadata")

    def get_digest(self) -> Digest:
        """
        .. versionadded:: 0.36.0
            Replaces the previous version of ``get_digest()``, now renamed to
            `get_raw_digest()`

        Retrieves the DANDI etag digest of the appropriate type for the asset:
        a dandi-etag digest for blob resources or a dandi-zarr-checksum for
        Zarr resources
        """
        algorithm = self.digest_type
        return Digest(algorithm=algorithm, value=self.get_raw_digest(algorithm))

    def get_content_url(
        self,
        regex: str = r".*",
        follow_redirects: Union[bool, int] = False,
        strip_query: bool = False,
    ) -> str:
        """
        Returns a URL for downloading the asset, found by inspecting the
        metadata; specifically, returns the first ``contentUrl`` that matches
        ``regex``.  Raises `NotFoundError` if the metadata does not contain a
        matching URL.

        If ``follow_redirects`` is `True`, a ``HEAD`` request is made to
        resolve any & all redirects before returning the URL.  If
        ``follow_redirects`` is an integer, at most that many redirects are
        followed.

        If ``strip_query`` is true, any query parameters are removed from the
        final URL before returning it.
        """
        url: str
        for url in self.get_raw_metadata().get("contentUrl", []):
            if re.search(regex, url):
                break
        else:
            raise NotFoundError(
                "No matching URL found in asset's contentUrl metadata field"
            )
        if follow_redirects is True:
            url = self.client.request(
                "HEAD", url, json_resp=False, allow_redirects=True
            ).url
        elif follow_redirects:
            for _ in range(follow_redirects):
                r = self.client.request(
                    "HEAD", url, json_resp=False, allow_redirects=False
                )
                if "Location" in r.headers:
                    url = r.headers["Location"]
                else:
                    break
        if strip_query:
            url = urlunparse(urlparse(url)._replace(query=""))
        return url

    def get_download_file_iter(
        self, chunk_size: int = MAX_CHUNK_SIZE
    ) -> Callable[[int], Iterator[bytes]]:
        """
        Returns a function that when called (optionally with an offset into the
        asset to start downloading at) returns a generator of chunks of the
        asset.

        :raises ValueError: if the asset is not backed by a blob
        """
        if self.asset_type is not AssetType.BLOB:
            raise ValueError(
                f"Cannot download asset {self} directly: asset is of type"
                f" {self.asset_type.name}, not BLOB"
            )

        url = self.base_download_url

        def downloader(start_at: int = 0) -> Iterator[bytes]:
            lgr.debug("Starting download from %s", url)
            headers = None
            if start_at > 0:
                headers = {"Range": f"bytes={start_at}-"}
            result = self.client.session.get(url, stream=True, headers=headers)
            # TODO: apparently we might need retries here as well etc
            # if result.status_code not in (200, 201):
            result.raise_for_status()
            for chunk in result.iter_content(chunk_size=chunk_size):
                if chunk:  # could be some "keep alive"?
                    yield chunk
            lgr.info("Asset %s successfully downloaded", self.identifier)

        return downloader

    def download(
        self, filepath: Union[str, Path], chunk_size: int = MAX_CHUNK_SIZE
    ) -> None:
        """
        Download the asset to ``filepath``.  Blocks until the download is
        complete.

        :raises ValueError: if the asset is not backed by a blob
        """
        downloader = self.get_download_file_iter(chunk_size=chunk_size)
        with open(filepath, "wb") as fp:
            for chunk in downloader(0):
                fp.write(chunk)

    @property
    @abstractmethod
    def asset_type(self) -> AssetType:
        """
        .. versionadded:: 0.36.0

        The type of the asset's underlying data
        """
        ...

    @property
    def digest_type(self) -> models.DigestType:
        """
        .. versionadded:: 0.36.0

        The primary digest algorithm used by Dandi Archive for the asset,
        determined based on its underlying data: dandi-etag for blob resources,
        dandi-zarr-checksum for Zarr resources
        """
        if self.asset_type is AssetType.ZARR:
            return models.DigestType.dandi_zarr_checksum
        else:
            return models.DigestType.dandi_etag


class BaseRemoteBlobAsset(BaseRemoteAsset):
    """
    .. versionadded:: 0.36.0

    A `BaseRemoteAsset` whose actual data is a blob resource
    """

    #: The ID of the underlying blob resource
    blob: str

    @property
    def asset_type(self) -> AssetType:
        """
        .. versionadded:: 0.36.0

        The type of the asset's underlying data
        """
        return AssetType.BLOB


class BaseRemoteZarrAsset(BaseRemoteAsset):
    """
    .. versionadded:: 0.36.0

    A `BaseRemoteAsset` whose actual data is a Zarr resource
    """

    #: The ID of the underlying Zarr resource
    zarr: str

    @property
    def asset_type(self) -> AssetType:
        """
        .. versionadded:: 0.36.0

        The type of the asset's underlying data
        """
        return AssetType.ZARR

    @property
    def filetree(self) -> "RemoteZarrEntry":
        """
        The `RemoteZarrEntry` for the root of the hierarchy of files within the
        Zarr
        """
        return RemoteZarrEntry(
            client=self.client, zarr_id=self.zarr, parts=(), _known_dir=True
        )

    def iterfiles(self, include_dirs: bool = False) -> Iterator["RemoteZarrEntry"]:
        """
        Returns a generator of all `RemoteZarrEntry`\\s within the Zarr.  By
        default, only instances for files are produced, unless ``include_dirs``
        is true.
        """
        return self.filetree.iterfiles(include_dirs=include_dirs)

    def rmfiles(
        self, files: Iterable["RemoteZarrEntry"], reingest: bool = True
    ) -> None:
        """
        Delete one or more files from the Zarr.

        If ``reingest`` is true, after performing the deletion, the client
        triggers a recalculation of the Zarr's checksum and waits for it to
        complete.
        """
        # Don't bother checking that the entries are actually files or even
        # belong to this Zarr, as if they're not, the server will return an
        # error anyway.
        for entries in chunked(files, ZARR_DELETE_BATCH_SIZE):
            self.client.delete(
                f"/zarr/{self.zarr}/files/",
                json=[{"path": str(e)} for e in entries],
            )
        if reingest:
            self.client.post(f"/zarr/{self.zarr}/ingest/")
            while True:
                sleep(2)
                r = self.client.get(f"/zarr/{self.zarr}/")
                if r["status"] == "Complete":
                    break


class RemoteAsset(BaseRemoteAsset):
    """
    Subclass of `BaseRemoteAsset` that includes information about the Dandiset
    to which the asset belongs.

    This is an abstract class; its concrete subclasses are `RemoteBlobAsset`
    (for assets backed by blobs) and `RemoteZarrAsset` (for assets backed by
    Zarrs).

    This class should not be instantiated by end-users directly.  Instead,
    instances should be retrieved from the appropriate attributes & methods of
    `RemoteDandiset`.
    """

    JSON_EXCLUDE = frozenset(["client", "dandiset_id", "version_id"])

    #: The identifier for the Dandiset to which the asset belongs
    dandiset_id: str
    #: The identifier for the version of the Dandiset to which the asset
    #: belongs
    version_id: str

    @classmethod
    def from_data(
        cls,
        dandiset: RemoteDandiset,
        data: Dict[str, Any],
        metadata: Optional[Dict[str, Any]] = None,
    ) -> "RemoteAsset":
        """
        Construct a `RemoteAsset` instance from a `RemoteDandiset`, a `dict` of
        raw data in the same format as returned by the API's pagination
        endpoints, and optional raw asset metadata.

        This is a low-level method that non-developers would normally only use
        when acquiring data using means outside of this library.
        """
        klass: Type[RemoteAsset]
        if data.get("blob") is not None:
            klass = RemoteBlobAsset
            if data.pop("zarr", None) is not None:
                raise ValueError("Asset data contains both `blob` and `zarr`'")
        elif data.get("zarr") is not None:
            klass = RemoteZarrAsset
            if data.pop("blob", None) is not None:
                raise ValueError("Asset data contains both `blob` and `zarr`'")
        else:
            raise ValueError("Asset data contains neither `blob` nor `zarr`")
        return klass(  # type: ignore[call-arg]
            client=dandiset.client,
            dandiset_id=dandiset.identifier,
            version_id=dandiset.version_id,
            **data,
            _metadata=metadata,
        )

    @property
    def api_path(self) -> str:
        """
        The path (relative to the base endpoint for a Dandi Archive API) at
        which API requests for interacting with the asset itself are made
        """
        return f"/dandisets/{self.dandiset_id}/versions/{self.version_id}/assets/{self.identifier}/"

    @property
    def api_url(self) -> str:
        """
        The URL at which API requests for interacting with the asset itself are
        made
        """
        return self.client.get_url(self.api_path)

    @property
    def download_url(self) -> str:
        """
        The URL from which the asset can be downloaded, including Dandiset
        identifiers (cf. `BaseRemoteAsset.base_download_url`)
        """
        return self.client.get_url(f"{self.api_path}download/")

    def set_metadata(self, metadata: models.Asset) -> None:
        """
        Set the metadata for the asset to the given value and update the
        `RemoteAsset` in place.
        """
        return self.set_raw_metadata(metadata.json_dict())

    @abstractmethod
    def set_raw_metadata(self, metadata: Dict[str, Any]) -> None:
        """
        Set the metadata for the asset on the server to the given value and
        update the `RemoteAsset` in place.
        """
        ...

    def delete(self) -> None:
        """Delete the asset"""
        self.client.delete(self.api_path)


class RemoteBlobAsset(RemoteAsset, BaseRemoteBlobAsset):
    """
    .. versionadded:: 0.36.0

    A `RemoteAsset` whose actual data is a blob resource
    """

    def set_raw_metadata(self, metadata: Dict[str, Any]) -> None:
        """
        Set the metadata for the asset on the server to the given value and
        update the `RemoteBlobAsset` in place.
        """
        data = self.client.put(
            self.api_path, json={"metadata": metadata, "blob_id": self.blob}
        )
        self.identifier = data["asset_id"]
        self.path = data["path"]
        self.size = int(data["size"])
        self.created = ensure_datetime(data["created"])
        self.modified = ensure_datetime(data["modified"])
        self._metadata = data["metadata"]


class RemoteZarrAsset(RemoteAsset, BaseRemoteZarrAsset):
    """
    .. versionadded:: 0.36.0

    A `RemoteAsset` whose actual data is a Zarr resource
    """

    def set_raw_metadata(self, metadata: Dict[str, Any]) -> None:
        """
        Set the metadata for the asset on the server to the given value and
        update the `RemoteZarrAsset` in place.
        """
        data = self.client.put(
            self.api_path, json={"metadata": metadata, "zarr_id": self.zarr}
        )
        self.identifier = data["asset_id"]
        self.path = data["path"]
        self.size = int(data["size"])
        self.created = ensure_datetime(data["created"])
        self.modified = ensure_datetime(data["modified"])
        self._metadata = data["metadata"]


class ZarrListing(BaseModel):
    """
    .. versionadded:: 0.36.0

    Information about a directory within a `RemoteZarrAsset`
    """

    #: API URLs for the listings of the directory's subdirectories
    directories: List[AnyHttpUrl]
    #: API URLs for downloading the files in the directory
    files: List[AnyHttpUrl]
    #: The checksums (MD5 or Dandi Zarr checksum, as appropriate) for the
    #: directory's entries, as a mapping from basenames to checksums
    checksums: Dict[str, str]
    #: The Dandi Zarr checksum for the directory
    checksum: str

    @property
    def dirnames(self) -> List[str]:
        """The basenames of the directory URLs in `directories`"""
        return [PurePosixPath(unquote(url.path or "")).name for url in self.directories]

    @property
    def filenames(self) -> List[str]:
        """The basenames of the file URLs in `files`"""
        return [PurePosixPath(unquote(url.path or "")).name for url in self.files]


@dataclass
class ZarrEntryStat:
    """
    .. versionadded:: 0.36.0

    Combined size & timestamp information for a file in a `RemoteZarrAsset`
    """

    #: The size of the file
    size: int
    #: The time at which the file was last modified
    modified: datetime


@dataclass
class RemoteZarrEntry(BasePath):
    """
    .. versionadded:: 0.36.0

    A file or directory within a `RemoteZarrAsset`.  Implements
    `~dandi.misctypes.BasePath`.
    """

    #: The `DandiAPIClient` instance used for API requests
    client: DandiAPIClient
    #: The ID of the Zarr backing the asset
    zarr_id: str
    _known_dir: Optional[bool] = field(default=None, compare=False, repr=False)
    _digest: Optional[Digest] = field(default=None, compare=False, repr=False)

    def _get_subpath(
        self, name: str, isdir: Optional[bool] = None, checksum: Optional[str] = None
    ) -> "RemoteZarrEntry":
        if not name or "/" in name:
            raise ValueError(f"Invalid path component: {name!r}")
        elif name == ".":
            return self
        elif name == "..":
            return self.parent
        else:
            if checksum is not None and isdir is not None:
                if isdir:
                    algorithm = models.DigestType.dandi_zarr_checksum
                else:
                    algorithm = models.DigestType.md5
                digest = Digest(algorithm=algorithm, value=checksum)
            else:
                digest = None
            return replace(
                self, parts=self.parts + (name,), _known_dir=isdir, _digest=digest
            )

    @property
    def parent(self) -> "RemoteZarrEntry":
        if self.is_root():
            return self
        else:
            return replace(
                self,
                parts=self.parts[:-1],
                _known_dir=True if self._known_dir is not None else None,
                _digest=None,
            )

    def _isdir(self) -> bool:
        if self._known_dir is not None:
            return self._known_dir
        elif self.is_root():
            try:
                self.client.get(f"/zarr/{self.zarr_id}/")
            except HTTP404Error:
                raise NotFoundError(f"No such Zarr: {self.zarr_id!r}")
            return True
        else:
            listing = self.parent.get_listing()
            if self.name in listing.dirnames:
                return True
            elif self.name in listing.filenames:
                return False
            else:
                raise NotFoundError(
                    f"No such entry {str(self)!r} in Zarr {self.zarr_id!r}"
                )

    def exists(self) -> bool:
        try:
            self._isdir()
        except NotFoundError:
            return False
        else:
            return True

    def is_file(self) -> bool:
        try:
            return not self._isdir()
        except NotFoundError:
            return False

    def is_dir(self) -> bool:
        try:
            return self._isdir()
        except NotFoundError:
            return False

    def iterdir(self) -> Iterator["RemoteZarrEntry"]:
        listing = self.get_listing()
        for name in listing.dirnames:
            if name == "." or name == "..":
                continue
            yield self._get_subpath(name, isdir=True, checksum=listing.checksums[name])
        for name in listing.filenames:
            yield self._get_subpath(name, isdir=False, checksum=listing.checksums[name])

    def get_digest(self) -> Digest:
        """
        Retrieve the DANDI etag digest for the entry.  If the entry is a
        directory, the algorithm will be the Dandi Zarr checksum algorithm; if
        it is a file, it will be MD5.

        :raises NotFoundError: if the path does not exist in the Zarr asset
        """
        if self._digest is not None:
            # `self` was returned by `parent.iterdir()`, which iterated over a
            # ZarrListing and thus had access to the checksum, which it stored
            # on `self`; use that value.
            return self._digest
        if self.is_root():
            algorithm = models.DigestType.dandi_zarr_checksum
            value = self.get_listing().checksum
        else:
            listing = self.parent.get_listing()
            if self.name in listing.dirnames:
                algorithm = models.DigestType.dandi_zarr_checksum
            elif self.name in listing.filenames:
                algorithm = models.DigestType.md5
            else:
                raise NotFoundError(
                    f"No such entry {str(self)!r} in Zarr {self.zarr_id!r}"
                )
            value = listing.checksums[self.name]
        return Digest(algorithm=algorithm, value=value)

    @property
    def size(self) -> int:
        """
        The size of the entry, which must be a file

        :raises NotFoundError: if the path does not exist in the Zarr asset
        :raises ValueError: if the path is a directory
        """
        return self.stat().size

    @property
    def modified(self) -> datetime:
        """
        The time at which the entry (which must be a file) was last modified

        :raises NotFoundError: if the path does not exist in the Zarr asset
        :raises ValueError: if the path is a directory
        """
        return self.stat().modified

    def stat(self) -> ZarrEntryStat:
        """
        Return combined size & timestamp information for the entry, which must
        be a file

        :raises NotFoundError: if the path does not exist in the Zarr asset
        :raises ValueError: if the path is a directory
        """
        if not self.is_file():
            raise ValueError("Cannot stat directories in Zarrs")
        try:
            r = self.client.request(
                "HEAD",
                f"/zarr/{self.zarr_id}.zarr/{'/'.join(self.parts)}",
                json_resp=False,
                allow_redirects=True,
            )
        except HTTP404Error:
            raise NotFoundError(
                f"{str(self)!r} in Zarr {self.zarr_id!r} does not exist"
            )
        return ZarrEntryStat(
            size=int(r.headers["Content-Length"]),
            modified=dateutil.parser.parse(r.headers["Last-Modified"]),
        )

    def get_listing(self) -> ZarrListing:
        """
        Return the `ZarrListing` for the entry, which must be a directory

        :raises NotFoundError:
            if the path does not exist in the Zarr asset or is not a directory
        """
        path = "".join(p + "/" for p in self.parts)
        try:
            r = self.client.get(f"/zarr/{self.zarr_id}.zarr/{path}")
        except HTTP404Error:
            raise NotFoundError(
                f"{str(self)!r} in Zarr {self.zarr_id!r} does not exist or is"
                " not a directory"
            )
        return ZarrListing.parse_obj(r)

    def get_download_file_iter(
        self, chunk_size: int = MAX_CHUNK_SIZE
    ) -> Callable[[int], Iterator[bytes]]:
        """
        Returns a function that when called (optionally with an offset into the
        file to start downloading at) returns a generator of chunks of the file
        """
        if not self.is_file():
            raise RuntimeError(
                f"{str(self)!r} in Zarr {self.zarr_id!r} does not exist or"
                " is not a file"
            )

        url = self.client.get_url(f"/zarr/{self.zarr_id}.zarr/{'/'.join(self.parts)}")

        def downloader(start_at: int = 0) -> Iterator[bytes]:
            lgr.debug("Starting download from %s", url)
            headers = None
            if start_at > 0:
                headers = {"Range": f"bytes={start_at}-"}
            result = self.client.session.get(url, stream=True, headers=headers)
            # TODO: apparently we might need retries here as well etc
            # if result.status_code not in (200, 201):
            result.raise_for_status()
            for chunk in result.iter_content(chunk_size=chunk_size):
                if chunk:  # could be some "keep alive"?
                    yield chunk
            lgr.info("File %s in Zarr %s successfully downloaded", self, self.zarr_id)

        return downloader

    def iterfiles(self, include_dirs: bool = False) -> Iterator["RemoteZarrEntry"]:
        """
        Returns a generator of all `RemoteZarrEntry`\\s under the directory.
        By default, only instances for files are produced, unless
        ``include_dirs`` is true.
        """
        dirs = deque([self])
        while dirs:
            d = dirs.popleft()
            try:
                subpaths = list(d.iterdir())
            except NotFoundError:
                if d.is_root():
                    # Empty zarr; can happen if an upload is interrupted during
                    # the initial batch
                    return
                else:
                    raise
            for p in subpaths:
                if p.is_dir():
                    dirs.append(p)
                    if include_dirs:
                        yield p
                else:
                    yield p

    def clear_cache(self) -> None:
        """Delete any locally-cached data about the entry"""
        self._known_dir = None
        self._digest = None



"""
This module provides functionality for parsing URLs and other resource
identifiers for Dandisets & assets on Dandi Archive servers and for fetching
the objects to which the URLs refer.  See :ref:`resource_ids` for a list of
accepted URL formats.

Basic operation begins by calling `parse_dandi_url()` on a URL in order to
acquire a `ParsedDandiURL` instance, which can then be used to obtain the
Dandiset and/or assets specified in the URL.  Call an instance's
`~ParsedDandiURL.get_dandiset()` and/or `~ParsedDandiURL.get_assets()` to get
the assets, passing in a `~dandi.dandiapi.DandiAPIClient` for the appropriate
Dandi Archive API instance; an unauthenticated client pointing to the correct
instance can be acquired via the `~ParsedDandiURL.get_client()` method.  As a
convenience, one can acquire a client, the Dandiset, and an iterator of all
assets by using the `~ParsedDandiAPI.navigate()` context manager like so:

.. code:: python

    parsed_url = parse_dandi_url("https://...")
    with parsed_url.navigate() as (client, dandiset, assets):
        ...
    # The client's session is closed when the context manager exits.

As a further convenience, a URL can be parsed and navigated in one fell swoop
using the `navigate_url()` function.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from contextlib import contextmanager
import re
from time import sleep
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, cast
from urllib.parse import unquote as urlunquote

from pydantic import AnyHttpUrl, BaseModel, parse_obj_as, validator
import requests

from . import get_logger
from .consts import (
    DANDISET_ID_REGEX,
    PUBLISHED_VERSION_REGEX,
    RETRY_STATUSES,
    VERSION_REGEX,
    known_instances,
)
from .dandiapi import BaseRemoteAsset, DandiAPIClient, RemoteDandiset
from .exceptions import FailedToConnectError, NotFoundError, UnknownURLError
from .utils import get_instance

lgr = get_logger()


class ParsedDandiURL(ABC, BaseModel):
    """
    Parsed representation of a URL pointing to a Dandi Archive resource
    (Dandiset or asset(s)).  Subclasses must implement `get_assets()`.

    Most methods take a ``client: DandiAPIClient`` argument, which must be a
    `~dandi.dandiapi.DandiAPIClient` object for querying `api_url` (This is not
    checked).  Such a client instance can be obtained by calling
    `get_client()`, or an appropriate pre-existing client instance can be
    passed instead."""

    #: The base URL of the Dandi API service, without a trailing slash
    api_url: AnyHttpUrl
    #: The ID of the Dandiset given in the URL
    dandiset_id: Optional[str]
    #: The version of the Dandiset, if specified.  If this is not set, the
    #: version will be defaulted using the rules described under
    #: `DandiAPIClient.get_dandiset()`.
    version_id: Optional[str] = None

    @validator("api_url")
    def _validate_api_url(cls, v: AnyHttpUrl) -> AnyHttpUrl:
        return cast(AnyHttpUrl, parse_obj_as(AnyHttpUrl, v.rstrip("/")))

    def get_client(self) -> DandiAPIClient:
        """
        Returns an unauthenticated `~dandi.dandiapi.DandiAPIClient` for
        `api_url`
        """
        return DandiAPIClient(self.api_url)

    def get_dandiset(
        self, client: DandiAPIClient, lazy: bool = True
    ) -> Optional[RemoteDandiset]:
        """
        Returns information about the specified (or default) version of the
        specified Dandiset.  Returns `None` if the URL did not contain a
        Dandiset identifier.

        If ``lazy`` is true, a "lazy" `RemoteDandiset` instance is returned,
        with no requests made until any data is actually required.
        """
        if self.dandiset_id is not None:
            return client.get_dandiset(self.dandiset_id, self.version_id, lazy=lazy)
        else:
            return None

    @abstractmethod
    def get_assets(
        self, client: DandiAPIClient, order: Optional[str] = None, strict: bool = False
    ) -> Iterator[BaseRemoteAsset]:
        """
        Returns an iterator of asset structures for the assets referred to by
        or associated with the URL.  For a URL pointing to just a Dandiset,
        this is the set of all assets in the given or default version of the
        Dandiset.  For a URL that specifies a specific asset or collection of
        assets in a Dandiset, this is all of those assets.

        When multiple assets are returned, they can be sorted by a given field
        by passing the name of that field as the ``order`` parameter.  The
        accepted field names are ``"created"``, ``"modified"``, and ``"path"``.
        Prepend a hyphen to the field name to reverse the sort order.

        If ``strict`` is true, then fetching assets for a URL that refers to a
        nonexistent resource will raise a `NotFoundError`; if it is false, the
        method will instead return an empty iterator.
        """
        ...

    def get_asset_ids(self, client: DandiAPIClient) -> Iterator[str]:
        """
        Returns an iterator of IDs of the assets referred to by or associated
        with the URL
        """
        for a in self.get_assets(client):
            yield a.identifier

    @contextmanager
    def navigate(
        self, *, strict: bool = False, authenticate: Optional[bool] = None
    ) -> Iterator[
        Tuple[DandiAPIClient, Optional[RemoteDandiset], Iterable[BaseRemoteAsset]]
    ]:
        """
        A context manager that returns a triple of a
        `~dandi.dandiapi.DandiAPIClient` (with an open session that is closed
        when the context manager closes), the return value of `get_dandiset()`,
        and the return value of `get_assets()`.

        If ``strict`` is true, then `get_dandiset()` is called with
        ``lazy=False`` and `get_assets()` is called with ``strict=True``; if
        ``strict`` is false, the opposite occurs.

        If ``authenticate`` is true, then
        `~dandi.dandiapi.DandiAPIClient.dandi_authenticate()` will be called on
        the client before returning it.  If it is `None` (the default), the
        method will only be called if the URL requires authentication (e.g., if
        the resource(s) are embargoed).

        .. versionchanged:: 0.35.0
            ``authenticate`` added
        """
        with self.get_client() as client:
            if authenticate:
                client.dandi_authenticate()
            try:
                dandiset = self.get_dandiset(client, lazy=not strict)
            except requests.HTTPError as e:
                if e.response.status_code == 401 and authenticate is not False:
                    lgr.info("Resource requires authentication; authenticating ...")
                    client.dandi_authenticate()
                    dandiset = self.get_dandiset(client, lazy=not strict)
                else:
                    raise
            assets = self.get_assets(client, strict=strict)
            yield (client, dandiset, assets)


class DandisetURL(ParsedDandiURL):
    """
    Parsed from a URL that only refers to a Dandiset (possibly with a version)
    """

    def get_assets(
        self, client: DandiAPIClient, order: Optional[str] = None, strict: bool = False
    ) -> Iterator[BaseRemoteAsset]:
        """Returns all assets in the Dandiset"""
        with _maybe_strict(strict):
            d = self.get_dandiset(client, lazy=not strict)
            assert d is not None
            yield from d.get_assets(order=order)


class SingleAssetURL(ParsedDandiURL):
    """Superclass for parsed URLs that refer to a single asset"""

    pass


class MultiAssetURL(ParsedDandiURL):
    """Superclass for parsed URLs that refer to multiple assets"""

    path: str


class BaseAssetIDURL(SingleAssetURL):
    """
    Parsed from a URL that refers to an asset by ID and does not include the
    Dandiset ID
    """

    asset_id: str

    def get_assets(
        self, client: DandiAPIClient, order: Optional[str] = None, strict: bool = False
    ) -> Iterator[BaseRemoteAsset]:
        """
        Yields the asset with the given ID.  If the asset does not exist, then
        a `NotFoundError` is raised if ``strict`` is true, and nothing is
        yielded if ``strict`` is false.
        """
        with _maybe_strict(strict):
            yield client.get_asset(self.asset_id)

    def get_asset_ids(self, client: DandiAPIClient) -> Iterator[str]:
        """Yields the ID of the asset (regardless of whether it exists)"""
        yield self.asset_id

    @contextmanager
    def navigate(
        self, *, strict: bool = False, authenticate: Optional[bool] = None
    ) -> Iterator[
        Tuple[DandiAPIClient, Optional[RemoteDandiset], Iterable[BaseRemoteAsset]]
    ]:
        with self.get_client() as client:
            if authenticate:
                client.dandi_authenticate()
            dandiset = self.get_dandiset(client, lazy=not strict)
            try:
                assets = list(self.get_assets(client, strict=strict))
            except requests.HTTPError as e:
                if e.response.status_code == 401 and authenticate is not False:
                    lgr.info("Resource requires authentication; authenticating ...")
                    client.dandi_authenticate()
                    assets = list(self.get_assets(client, strict=strict))
                else:
                    raise
            yield (client, dandiset, assets)


class AssetIDURL(SingleAssetURL):
    """
    Parsed from a URL that refers to an asset by ID and includes the Dandiset ID
    """

    asset_id: str

    def get_assets(
        self, client: DandiAPIClient, order: Optional[str] = None, strict: bool = False
    ) -> Iterator[BaseRemoteAsset]:
        """
        Yields the asset with the given ID.  If the Dandiset or asset does not
        exist, then a `NotFoundError` is raised if ``strict`` is true, and
        nothing is yielded if ``strict`` is false.
        """
        with _maybe_strict(strict):
            d = self.get_dandiset(client, lazy=not strict)
            assert d is not None
            yield d.get_asset(self.asset_id)

    def get_asset_ids(self, client: DandiAPIClient) -> Iterator[str]:
        """Yields the ID of the asset (regardless of whether it exists)"""
        yield self.asset_id


class AssetPathPrefixURL(MultiAssetURL):
    """
    Parsed from a URL that refers to a collection of assets by path prefix
    """

    def get_assets(
        self, client: DandiAPIClient, order: Optional[str] = None, strict: bool = False
    ) -> Iterator[BaseRemoteAsset]:
        """Returns the assets whose paths start with `path`"""
        with _maybe_strict(strict):
            d = self.get_dandiset(client, lazy=not strict)
            assert d is not None
            yield from d.get_assets_with_path_prefix(self.path, order=order)


class AssetItemURL(SingleAssetURL):
    """Parsed from a URL that refers to a specific asset by path"""

    path: str

    def get_assets(
        self, client: DandiAPIClient, order: Optional[str] = None, strict: bool = False
    ) -> Iterator[BaseRemoteAsset]:
        """
        Yields the asset whose path equals `path`.  If there is no such asset:

        - If ``strict`` is true, a `NotFoundError` is raised.
        - If ``strict`` is false and the path happens to be the path to an
          asset directory, a `ValueError` is raised indicating that the user
          left off a trailing slash.
        - Otherwise, nothing is yielded.
        """
        try:
            dandiset = self.get_dandiset(client, lazy=not strict)
            assert dandiset is not None
            # Force evaluation of the version here instead of when
            # get_asset_by_path() is called so we don't get nonexistent
            # dandisets with unspecified versions mixed up with nonexistent
            # assets.
            dandiset.version_id
        except NotFoundError:
            if strict:
                raise
            else:
                return
        try:
            yield dandiset.get_asset_by_path(self.path)
        except NotFoundError:
            if strict:
                raise
            try:
                next(dandiset.get_assets_with_path_prefix(self.path + "/"))
            except NotFoundError:
                # Dandiset has explicit version that doesn't exist.
                return
            except StopIteration:
                pass
            else:
                raise ValueError(
                    f"Asset path {self.path!r} points to a directory but lacks trailing /"
                )


class AssetFolderURL(MultiAssetURL):
    """
    Parsed from a URL that refers to a collection of assets by folder path
    """

    path: str

    def get_assets(
        self, client: DandiAPIClient, order: Optional[str] = None, strict: bool = False
    ) -> Iterator[BaseRemoteAsset]:
        """
        Returns all assets under the folder at `path`.  Yields nothing if the
        folder does not exist.
        """
        path = self.path
        if not path.endswith("/"):
            path += "/"
        with _maybe_strict(strict):
            d = self.get_dandiset(client, lazy=not strict)
            assert d is not None
            yield from d.get_assets_with_path_prefix(path, order=order)


@contextmanager
def _maybe_strict(strict: bool) -> Iterator[None]:
    try:
        yield
    except NotFoundError:
        if strict:
            raise


@contextmanager
def navigate_url(
    url: str, *, strict: bool = False, authenticate: Optional[bool] = None
) -> Iterator[
    Tuple[DandiAPIClient, Optional[RemoteDandiset], Iterable[BaseRemoteAsset]]
]:
    """
    A context manager that takes a URL pointing to a DANDI Archive and
    returns a triple of a `~dandi.dandiapi.DandiAPIClient` (with an open
    session that is closed when the context manager closes), the Dandiset
    identified in the URL (if any), and the assets specified by the URL (or, if
    no specific assets were specified, all assets in the Dandiset).

    .. versionchanged:: 0.35.0
        ``authenticate`` added

    :param str url: URL which might point to a Dandiset, folder, or asset(s)
    :param bool strict:
        If true, then `get_dandiset()` is called with ``lazy=False`` and
        `get_assets()` is called with ``strict=True``; if false, the opposite
        occurs.
    :param Optional[bool] authenticate:
        If true, then `~dandi.dandiapi.DandiAPIClient.dandi_authenticate()`
        will be called on the client before returning it.  If `None` (the
        default), the method will only be called if the URL requires
        authentication (e.g., if the resource(s) are embargoed).
    :returns: Context manager that yields a ``(client, dandiset, assets)``
        tuple; ``client`` will have a session established for the duration of
        the context
    """
    parsed_url = parse_dandi_url(url)
    with parsed_url.navigate(strict=strict, authenticate=authenticate) as (
        client,
        dandiset,
        assets,
    ):
        yield (client, dandiset, assets)


class _dandi_url_parser:
    # Defining as a class with all the attributes to not leak all the variables
    # etc into module space, and later we might end up with classes for those
    # anyways
    dandiset_id_grp = f"(?P<dandiset_id>{DANDISET_ID_REGEX})"
    # Should absorb port and "api/":
    server_grp = "(?P<server>(?P<protocol>https?)://(?P<hostname>[^/]+)/(api/)?)"
    known_urls: List[Tuple[re.Pattern[str], Dict[str, Any], str]] = [
        # List of (regex, settings, display string) triples
        #
        # Settings:
        # - handle_redirect:
        #   - 'pass' - would continue with original url if no redirect happen
        #   - 'only' - would interrupt if no redirection happens
        # - rewrite:
        #   - callable -- which would rewrite that "URI"
        # - map_instance
        #
        # Those we first redirect and then handle the redirected URL
        # TODO: Later should better conform to our API, so we could allow
        #       for not only "dandiarchive.org" URLs
        (
            re.compile(
                fr"DANDI:{DANDISET_ID_REGEX}(?:/{PUBLISHED_VERSION_REGEX})?",
                flags=re.I,
            ),
            {"rewrite": lambda x: "https://identifiers.org/" + x.lower()},
            "DANDI:<dandiset id>[/<version id>]",
        ),
        (
            re.compile(r"https?://dandiarchive\.org/.*"),
            {"handle_redirect": "pass"},
            "https://dandiarchive.org/...",
        ),
        (
            re.compile(
                fr"https?://identifiers\.org/DANDI:{DANDISET_ID_REGEX}"
                fr"(?:/{PUBLISHED_VERSION_REGEX})?",
                flags=re.I,
            ),
            {"handle_redirect": "pass"},
            "https://identifiers.org/DANDI:<dandiset id>[/<version id>]",
        ),
        (
            re.compile(
                rf"{server_grp}(#/)?(?P<asset_type>dandiset)/{dandiset_id_grp}"
                rf"(/(?P<version>{VERSION_REGEX}))?"
                r"(/(files(\?location=(?P<location>.*)?)?)?)?"
            ),
            {},
            "https://<server>[/api]/[#/]dandiset/<dandiset id>[/<version>]"
            "[/files[?location=<path>]]",
        ),
        # PRs are also on netlify - so above takes precedence. TODO: make more
        # specific?
        (
            re.compile(r"https?://[^/]*dandiarchive-org\.netlify\.app/.*"),
            {"map_instance": "dandi"},
            "https://*dandiarchive-org.netflify.app/...",
        ),
        # Direct urls to our new API
        (
            re.compile(
                rf"{server_grp}(?P<asset_type>dandiset)s/{dandiset_id_grp}"
                rf"(/(versions(/(?P<version>{VERSION_REGEX}))?/?)?)?"
            ),
            {},
            "https://<server>[/api]/dandisets/<dandiset id>[/versions[/<version>]]",
        ),
        (
            re.compile(
                rf"{server_grp}(?P<asset_type>asset)s/(?P<asset_id>[^?/]+)(/(download/?)?)?"
            ),
            {},
            "https://<server>[/api]/assets/<asset id>[/download]",
        ),
        (
            re.compile(
                rf"{server_grp}(?P<asset_type>dandiset)s/{dandiset_id_grp}"
                rf"/versions/(?P<version>{VERSION_REGEX})"
                r"/assets/(?P<asset_id>[^?/]+)(/(download/?)?)?"
            ),
            {},
            "https://<server>[/api]/dandisets/<dandiset id>/versions/<version>"
            "/assets/<asset id>[/download]",
        ),
        (
            re.compile(
                rf"{server_grp}(?P<asset_type>dandiset)s/{dandiset_id_grp}"
                rf"/versions/(?P<version>{VERSION_REGEX})"
                r"/assets/\?path=(?P<path>[^&]+)",
            ),
            {},
            "https://<server>[/api]/dandisets/<dandiset id>/versions/<version>"
            "/assets/?path=<path>",
        ),
        # ad-hoc explicitly pointing within URL to the specific instance to use
        # and otherwise really simple:
        # dandi://INSTANCE/DANDISET_ID[@VERSION][/PATH]
        # For now to not be promoted to the users, and primarily for internal
        # use
        (
            re.compile(
                rf"dandi://(?P<instance_name>[-\w._]+)"
                rf"/{dandiset_id_grp}"
                rf"(@(?P<version>{VERSION_REGEX}))?"
                rf"(/(?P<location>.*)?)?"
            ),
            {},
            "dandi://<instance name>/<dandiset id>[@<version>][/<path>]",
        ),
        # https://deploy-preview-341--gui-dandiarchive-org.netlify.app/#/dandiset/000006/draft
        # (no API yet)
        (
            re.compile(r"https?://.*"),
            {"handle_redirect": "only"},
            "https://<server>/...",
        ),
    ]
    known_patterns = "Accepted resource identifier patterns:" + "\n - ".join(
        [""] + [display for _, _, display in known_urls]
    )
    map_to = {}
    for (gui, redirector, api) in known_instances.values():
        if api:
            for h in (gui, redirector):
                if h:
                    map_to[h] = api

    @classmethod
    def parse(cls, url: str, *, map_instance: bool = True) -> ParsedDandiURL:
        """
        Parse a Dandi Archive URL and return a `ParsedDandiURL` instance.  See
        :ref:`resource_ids` for the supported URL formats.

        :raises UnknownURLError: if the URL is not one of the above
        """

        lgr.debug("Parsing url %s", url)

        # Loop through known url regexes and stop as soon as one is matching
        match = None
        for regex, settings, _ in cls.known_urls:
            match = regex.fullmatch(url)
            if not match:
                continue
            groups = match.groupdict()
            lgr.log(5, "Matched %r into %s", url, groups)
            rewrite = settings.get("rewrite", False)
            handle_redirect = settings.get("handle_redirect", False)
            if rewrite:
                assert not handle_redirect
                assert not settings.get("map_instance")
                new_url = rewrite(url)
                return cls.parse(new_url)
            elif handle_redirect:
                assert handle_redirect in ("pass", "only")
                new_url = cls.follow_redirect(url)
                if new_url != url:
                    return cls.parse(new_url)
                if handle_redirect == "pass":
                    # We used to issue warning in such cases, but may be it got implemented
                    # now via reverse proxy and we had added a new regex? let's just
                    # continue with a debug msg
                    lgr.debug("Redirection did not happen for %s", url)
                else:
                    raise RuntimeError(
                        f"{url} did not redirect to another location which dandi client would"
                        f" know how to handle."
                    )
            elif settings.get("map_instance"):
                if map_instance:
                    parsed_url = cls.parse(url, map_instance=False)
                    if settings["map_instance"] not in known_instances:
                        raise ValueError(
                            "Unknown instance {}. Known are: {}".format(
                                settings["map_instance"], ", ".join(known_instances)
                            )
                        )
                    known_instance = get_instance(settings["map_instance"])
                    assert known_instance.api is not None
                    parsed_url.api_url = cast(
                        AnyHttpUrl, parse_obj_as(AnyHttpUrl, known_instance.api)
                    )
                continue  # in this run we ignore and match further
            elif "instance_name" in groups:
                try:
                    known_instance = get_instance(groups["instance_name"])
                except KeyError:
                    raise UnknownURLError(
                        f"Unknown instance {groups['instance_name']!r}.  Valid instances: "
                        + ", ".join(sorted(known_instances))
                    )
                assert known_instance.api  # must be defined
                groups["server"] = known_instance.api
                # could be overloaded later depending if location is provided
                groups["asset_type"] = "dandiset"
                break
            else:
                break
        if not match:
            # TODO: may be make use of etelemetry and report if newer client
            # which might know is available?
            raise UnknownURLError(
                f"We do not know how to map URL {url} to our servers.\n"
                f"{cls.known_patterns}"
            )

        url_server = groups["server"].rstrip("/")
        if re.fullmatch(
            r"https?://deploy-preview-.*--gui-dandiarchive-org\.netlify\.app",
            url_server,
        ):
            url_server = "https://gui-staging.dandiarchive.org"
        server = cls.map_to.get(url_server, url_server)
        # asset_type = groups.get("asset_type")
        dandiset_id = groups.get("dandiset_id")
        version_id = groups.get("version")
        location = groups.get("location")
        asset_id = groups.get("asset_id")
        path = groups.get("path")
        if location:
            location = urlunquote(location)
            # ATM carries leading '/' which IMHO is not needed/misguiding
            # somewhat, so I will just strip it
            location = location.lstrip("/")
        # if location is not degenerate -- it would be a folder or a file
        if location:
            if location.endswith("/"):
                parsed_url = AssetFolderURL(
                    api_url=server,
                    dandiset_id=dandiset_id,
                    version_id=version_id,
                    path=location,
                )
            else:
                parsed_url = AssetItemURL(
                    api_url=server,
                    dandiset_id=dandiset_id,
                    version_id=version_id,
                    path=location,
                )
        elif asset_id:
            if dandiset_id is None:
                parsed_url = BaseAssetIDURL(api_url=server, asset_id=asset_id)
            else:
                parsed_url = AssetIDURL(
                    api_url=server,
                    dandiset_id=dandiset_id,
                    version_id=version_id,
                    asset_id=asset_id,
                )
        elif path:
            parsed_url = AssetPathPrefixURL(
                api_url=server,
                dandiset_id=dandiset_id,
                version_id=version_id,
                path=path,
            )
        else:
            parsed_url = DandisetURL(
                api_url=server,
                dandiset_id=dandiset_id,
                version_id=version_id,
            )
        lgr.debug("Parsed into %r", parsed_url)
        return parsed_url

    @staticmethod
    def follow_redirect(url: str) -> str:
        """
        Resolve the given URL by following all redirects.

        :raises NotFoundError: if a 404 response is returned
        :raises FailedToConnectError: if a response other than 200, 400, 404,
            or one of the statuses in `~dandi.consts.RETRY_STATUSES` is returned
        """
        i = 0
        while True:
            r = requests.head(url, allow_redirects=True)
            if r.status_code in RETRY_STATUSES and i < 4:
                delay = 0.1 * 10 ** i
                lgr.warning(
                    "HEAD request to %s returned %d; sleeping for %f seconds and then retrying...",
                    url,
                    r.status_code,
                    delay,
                )
                sleep(delay)
                i += 1
                continue
            elif r.status_code == 404:
                raise NotFoundError(url)
            elif r.status_code != 200:
                raise FailedToConnectError(
                    f"Response for getting {url} to redirect returned {r.status_code}."
                    f" Please verify that it is a URL related to dandiarchive and"
                    f" supported by dandi client"
                )
            elif r.url != url:
                return r.url
            assert isinstance(url, str)
            return url


# convenience binding
parse_dandi_url = _dandi_url_parser.parse
follow_redirect = _dandi_url_parser.follow_redirect



from operator import attrgetter
from pathlib import Path

from dandischema.models import get_schema_version
import numpy as np
import zarr

from .. import get_logger
from ..consts import ZARR_MIME_TYPE, dandiset_metadata_file
from ..dandiapi import AssetType, RemoteZarrAsset
from ..files import (
    DandisetMetadataFile,
    GenericAsset,
    NWBAsset,
    VideoAsset,
    ZarrAsset,
    dandi_file,
    find_dandi_files,
)

lgr = get_logger()


def test_find_dandi_files(tmp_path: Path) -> None:
    (tmp_path / dandiset_metadata_file).touch()
    (tmp_path / "sample01.zarr").mkdir()
    (tmp_path / "sample01.zarr" / "inner.nwb").touch()
    (tmp_path / "sample01.zarr" / "foo").touch()
    (tmp_path / "sample02.nwb").touch()
    (tmp_path / "foo").touch()
    (tmp_path / "bar.txt").touch()
    (tmp_path / "subdir").mkdir()
    (tmp_path / "subdir" / "sample03.nwb").touch()
    (tmp_path / "subdir" / "sample04.zarr").mkdir()
    (tmp_path / "subdir" / "sample04.zarr" / "inner2.nwb").touch()
    (tmp_path / "subdir" / "sample04.zarr" / "baz").touch()
    (tmp_path / "subdir" / "gnusto").touch()
    (tmp_path / "subdir" / "cleesh.txt").touch()
    (tmp_path / "empty.zarr").mkdir()
    (tmp_path / "glarch.mp4").touch()
    (tmp_path / ".ignored").touch()
    (tmp_path / ".ignored.dir").mkdir()
    (tmp_path / ".ignored.dir" / "ignored.nwb").touch()

    files = sorted(
        find_dandi_files(tmp_path, dandiset_path=tmp_path), key=attrgetter("filepath")
    )
    assert files == [
        VideoAsset(filepath=tmp_path / "glarch.mp4", path="glarch.mp4"),
        ZarrAsset(filepath=tmp_path / "sample01.zarr", path="sample01.zarr"),
        NWBAsset(filepath=tmp_path / "sample02.nwb", path="sample02.nwb"),
        NWBAsset(
            filepath=tmp_path / "subdir" / "sample03.nwb", path="subdir/sample03.nwb"
        ),
        ZarrAsset(
            filepath=tmp_path / "subdir" / "sample04.zarr", path="subdir/sample04.zarr"
        ),
    ]

    files = sorted(
        find_dandi_files(tmp_path, dandiset_path=tmp_path, allow_all=True),
        key=attrgetter("filepath"),
    )
    assert files == [
        GenericAsset(filepath=tmp_path / "bar.txt", path="bar.txt"),
        DandisetMetadataFile(filepath=tmp_path / dandiset_metadata_file),
        GenericAsset(filepath=tmp_path / "foo", path="foo"),
        VideoAsset(filepath=tmp_path / "glarch.mp4", path="glarch.mp4"),
        ZarrAsset(filepath=tmp_path / "sample01.zarr", path="sample01.zarr"),
        NWBAsset(filepath=tmp_path / "sample02.nwb", path="sample02.nwb"),
        GenericAsset(
            filepath=tmp_path / "subdir" / "cleesh.txt", path="subdir/cleesh.txt"
        ),
        GenericAsset(filepath=tmp_path / "subdir" / "gnusto", path="subdir/gnusto"),
        NWBAsset(
            filepath=tmp_path / "subdir" / "sample03.nwb", path="subdir/sample03.nwb"
        ),
        ZarrAsset(
            filepath=tmp_path / "subdir" / "sample04.zarr", path="subdir/sample04.zarr"
        ),
    ]

    files = sorted(
        find_dandi_files(tmp_path, dandiset_path=tmp_path, include_metadata=True),
        key=attrgetter("filepath"),
    )
    assert files == [
        DandisetMetadataFile(filepath=tmp_path / dandiset_metadata_file),
        VideoAsset(filepath=tmp_path / "glarch.mp4", path="glarch.mp4"),
        ZarrAsset(filepath=tmp_path / "sample01.zarr", path="sample01.zarr"),
        NWBAsset(filepath=tmp_path / "sample02.nwb", path="sample02.nwb"),
        NWBAsset(
            filepath=tmp_path / "subdir" / "sample03.nwb", path="subdir/sample03.nwb"
        ),
        ZarrAsset(
            filepath=tmp_path / "subdir" / "sample04.zarr", path="subdir/sample04.zarr"
        ),
    ]


def test_validate_simple1(simple1_nwb):
    # this file should be ok as long as schema_version is specified
    errors = dandi_file(simple1_nwb).get_validation_errors(
        schema_version=get_schema_version()
    )
    assert not errors


def test_validate_simple1_no_subject(simple1_nwb):
    errors = dandi_file(simple1_nwb).get_validation_errors()
    assert errors == ["Subject is missing."]


def test_validate_simple2(simple2_nwb):
    # this file should be ok since a Subject is included
    errors = dandi_file(simple2_nwb).get_validation_errors()
    assert not errors


def test_validate_simple2_new(simple2_nwb):
    # this file should be ok
    errors = dandi_file(simple2_nwb).get_validation_errors(
        schema_version=get_schema_version()
    )
    assert not errors


def test_validate_simple3_no_subject_id(simple3_nwb):
    errors = dandi_file(simple3_nwb).get_validation_errors()
    assert errors == ["subject_id is missing."]


def test_validate_bogus(tmp_path):
    path = tmp_path / "wannabe.nwb"
    path.write_text("not really nwb")
    # intended to produce use-case for https://github.com/dandi/dandi-cli/issues/93
    # but it would be tricky, so it is more of a smoke test that
    # we do not crash
    errors = dandi_file(path).get_validation_errors()
    # ATM we would get 2 errors -- since could not be open in two places,
    # but that would be too rigid to test. Let's just see that we have expected errors
    assert any(e.startswith("Failed to inspect NWBFile") for e in errors)


def test_upload_zarr(new_dandiset, tmp_path):
    filepath = tmp_path / "example.zarr"
    zarr.save(filepath, np.arange(1000), np.arange(1000, 0, -1))
    zf = dandi_file(filepath)
    assert isinstance(zf, ZarrAsset)
    asset = zf.upload(new_dandiset.dandiset, {"description": "A test Zarr"})
    assert isinstance(asset, RemoteZarrAsset)
    assert asset.asset_type is AssetType.ZARR
    assert asset.path == "example.zarr"
    md = asset.get_raw_metadata()
    assert md["encodingFormat"] == ZARR_MIME_TYPE
    assert md["description"] == "A test Zarr"
    md["description"] = "A modified Zarr"
    asset.set_raw_metadata(md)
    md = asset.get_raw_metadata()
    assert md["description"] == "A modified Zarr"

    for file_src in [zf, asset]:
        lgr.debug("Traversing %s", type(file_src).__name__)
        entries = sorted(file_src.iterfiles(include_dirs=True), key=attrgetter("parts"))
        assert [str(e) for e in entries] == [
            ".zgroup",
            "arr_0",
            "arr_0/.zarray",
            "arr_0/0",
            "arr_1",
            "arr_1/.zarray",
            "arr_1/0",
        ]
        assert (file_src.filetree / ".zgroup").exists()
        assert (file_src.filetree / ".zgroup").is_file()
        assert not (file_src.filetree / ".zgroup").is_dir()
        assert (file_src.filetree / "arr_0").exists()
        assert not (file_src.filetree / "arr_0").is_file()
        assert (file_src.filetree / "arr_0").is_dir()
        assert not (file_src.filetree / "0").exists()
        assert not (file_src.filetree / "0").is_file()
        assert not (file_src.filetree / "0").is_dir()
        assert not (file_src.filetree / "arr_0" / ".zgroup").exists()
        assert not (file_src.filetree / "arr_0" / ".zgroup").is_file()
        assert not (file_src.filetree / "arr_0" / ".zgroup").is_dir()
        assert not (file_src.filetree / ".zgroup" / "0").exists()
        assert not (file_src.filetree / ".zgroup" / "0").is_file()
        assert not (file_src.filetree / ".zgroup" / "0").is_dir()
        assert not (file_src.filetree / "arr_2" / "0").exists()
        assert not (file_src.filetree / "arr_2" / "0").is_file()
        assert not (file_src.filetree / "arr_2" / "0").is_dir()



"""
.. versionadded:: 0.36.0

Miscellaneous public classes
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from fnmatch import fnmatchcase
from typing import Dict, Iterator, List, Tuple, TypeVar

from dandischema.models import DigestType


@dataclass
class Digest:
    """A computed digest for a file or directory"""

    #: The digest algorithm used
    algorithm: DigestType

    #: The digest itself
    value: str

    @classmethod
    def dandi_etag(cls, value: str) -> Digest:
        """
        Construct a `Digest` with the given value and a ``algorithm`` of
        ``DigestType.dandi_etag``
        """
        return cls(algorithm=DigestType.dandi_etag, value=value)

    @classmethod
    def dandi_zarr(cls, value: str) -> Digest:
        """
        Construct a `Digest` with the given value and a ``algorithm`` of
        ``DigestType.dandi_zarr_checksum``
        """
        return cls(algorithm=DigestType.dandi_zarr_checksum, value=value)

    def asdict(self) -> Dict[DigestType, str]:
        """
        Convert the instance to a single-item `dict` mapping the digest
        algorithm to the digest value
        """
        return {self.algorithm: self.value}


#: Placeholder digest used in some situations where a digest is required but
#: not actually relevant and would be too expensive to calculate
DUMMY_DIGEST = Digest(algorithm=DigestType.dandi_etag, value=32 * "d" + "-1")

P = TypeVar("P", bound="BasePath")


@dataclass  # type: ignore[misc]  # <https://github.com/python/mypy/issues/5374>
class BasePath(ABC):
    """
    An abstract base class for path-like objects that can be traversed with the
    ``/`` operator *à la* `pathlib.Path` (though, unlike `pathlib.Path`
    instances, "dividing" by another non-string path is not allowed).  All
    paths are treated as forward-slash-separated relative paths under an
    empty-name "root" path.
    """

    #: The path components of the object
    parts: Tuple[str, ...]

    def __str__(self) -> str:
        return "/".join(self.parts)

    @property
    def name(self) -> str:
        """
        The basename of the path object.  When the object represents the root
        of a path hierarchy, this is the empty string.
        """
        if self.is_root():
            return ""
        else:
            assert self.parts
            return self.parts[-1]

    @abstractmethod
    def _get_subpath(self: P, name: str) -> P:
        """
        Return the path immediately under the instance with the given name.  A
        name of ``"."`` should cause ``self`` to be returned, and a name of
        ``".."`` should cause ``self.parent`` to be returned.  An empty name or
        a name containing a forward slash should result in a `ValueError`.
        """
        ...

    def __truediv__(self: P, path: str) -> P:
        p = self
        for q in self._split_path(path):
            p = p._get_subpath(q)
        return p

    def joinpath(self: P, *paths: str) -> P:
        """
        Combine the path with each name or relative path in ``paths`` using the
        ``/`` operator
        """
        p = self
        for q in paths:
            p /= q
        return p

    @staticmethod
    def _split_path(path: str) -> Tuple[str, ...]:
        """Split a path into its path components"""
        if path.startswith("/"):
            raise ValueError(f"Absolute paths not allowed: {path!r}")
        return tuple(q for q in path.split("/") if q)

    def is_root(self) -> bool:
        """
        Returns true if this path object represents the root of its hierarchy
        """
        return self.parts == ()

    @property
    def root_path(self: P) -> P:
        """The root of the path object's hierarchy"""
        p = self
        while not p.is_root():
            p = p.parent
        return p

    @property
    @abstractmethod
    def parent(self: P) -> P:
        """
        The parent path of the object.  The parent of the root of a path
        hierarchy is itself.
        """
        ...

    @property
    def parents(self: P) -> Tuple[P, ...]:
        """
        A tuple of the path's ancestors, starting at the parent and going up to
        (and including) the root of the hierarchy
        """
        ps: List[P] = []
        p = self
        while not p.is_root():
            q = p.parent
            ps.append(q)
            p = q
        return tuple(ps)

    def with_name(self: P, name: str) -> P:
        """Equivalent to ``p.parent / name``"""
        return self.parent / name

    @property
    def suffix(self) -> str:
        """The final file extension of the basename, if any"""
        i = self.name.rfind(".")
        if 0 < i < len(self.name) - 1:
            return self.name[i:]
        else:
            return ""

    @property
    def suffixes(self) -> List[str]:
        """A list of the basename's file extensions"""
        if self.name.endswith("."):
            return []
        name = self.name.lstrip(".")
        return ["." + suffix for suffix in name.split(".")[1:]]

    @property
    def stem(self) -> str:
        """The basename without its final file extension, if any"""
        i = self.name.rfind(".")
        if 0 < i < len(self.name) - 1:
            return self.name[:i]
        else:
            return self.name

    def with_stem(self: P, stem: str) -> P:
        """Returns a new path with the stem changed"""
        return self.with_name(stem + self.suffix)

    def with_suffix(self: P, suffix: str) -> P:
        """Returns a new path with the final file extension changed"""
        if "/" in suffix or (suffix and not suffix.startswith(".")) or suffix == ".":
            raise ValueError(f"Invalid suffix: {suffix!r}")
        if not self.name:
            raise ValueError("Path has an empty name")
        if not self.suffix:
            name = self.name + suffix
        else:
            name = self.name[: -len(self.suffix)] + suffix
        return self.with_name(name)

    def match(self, pattern: str) -> bool:
        """Tests whether the path matches the given glob pattern"""
        patparts = self._split_path(pattern)
        if not patparts:
            raise ValueError("Empty pattern")
        if len(patparts) > len(self.parts):
            return False
        for part, pat in zip(reversed(self.parts), reversed(patparts)):
            if not fnmatchcase(part, pat):
                return False
        return True

    @abstractmethod
    def exists(self) -> bool:
        """True iff the resource at the given path exists"""
        ...

    @abstractmethod
    def is_file(self) -> bool:
        """True if the resource at the given path exists and is a file"""
        ...

    @abstractmethod
    def is_dir(self) -> bool:
        """True if the resource at the given path exists and is a directory"""
        ...

    @abstractmethod
    def iterdir(self: P) -> Iterator[P]:
        """
        Returns a generator of the paths under the instance, which must be a
        directory
        """
        ...

    @property
    @abstractmethod
    def size(self) -> int:
        """The size of the resource at the path"""
        ...



import click

from .base import (
    IntColonInt,
    devel_debug_option,
    devel_option,
    instance_option,
    map_to_click_exceptions,
)


@click.command()
@click.option(
    "-e",
    "--existing",
    type=click.Choice(["error", "skip", "force", "overwrite", "refresh"]),
    help="What to do if a file found existing on the server. 'skip' would skip"
    "the file, 'force' - force reupload, 'overwrite' - force upload if "
    "either size or modification time differs; 'refresh' - upload only if "
    "local modification time is ahead of the remote.",
    default="refresh",
    show_default=True,
)
@click.option(
    "-J",
    "--jobs",
    type=IntColonInt(),
    help=(
        "Number of assets to upload in parallel and, optionally, number of"
        " upload threads per asset  [default: 5:5]"
    ),
)
@click.option(
    "--sync", is_flag=True, help="Delete assets on the server that do not exist locally"
)
@click.option(
    "--validation",
    help="Data must pass validation before the upload.  Use of this option is highly discouraged.",
    type=click.Choice(["require", "skip", "ignore"]),
    default="require",
    show_default=True,
)
@click.argument("paths", nargs=-1)  # , type=click.Path(exists=True, dir_okay=False))
# &
# Development options:  Set DANDI_DEVEL for them to become available
#
# TODO: should always go to dandi for now
@instance_option()
@devel_option(
    "--allow-any-path",
    help="For development: allow DANDI 'unsupported' file types/paths",
    default=False,
    is_flag=True,
)
@devel_option(
    "--upload-dandiset-metadata",
    help="For development: do upload dandiset metadata",
    default=False,
    is_flag=True,
)
@devel_debug_option()
@map_to_click_exceptions
def upload(
    paths,
    jobs,
    sync,
    dandi_instance,
    existing="refresh",
    validation="require",
    # Development options should come as kwargs
    allow_any_path=False,
    upload_dandiset_metadata=False,
    devel_debug=False,
):
    """
    Upload Dandiset files to DANDI Archive.

    The target Dandiset to upload to must already be registered in the archive,
    and a `dandiset.yaml` file must exist in the common ancestor of the given
    paths (or the current directory, if no paths are specified) or a parent
    directory thereof.

    Local Dandiset should pass validation.  For that, the assets should first
    be organized using the `dandi organize` command.

    By default all .nwb, .zarr, and .ngff assets in the Dandiset (ignoring
    directories starting with a period) will be considered for the upload.  You
    can point to specific files you would like to validate and have uploaded.
    """
    from ..upload import upload

    if jobs is None:
        jobs = None
        jobs_per_file = None
    else:
        jobs, jobs_per_file = jobs

    upload(
        paths,
        existing=existing,
        validation=validation,
        dandi_instance=dandi_instance,
        allow_any_path=allow_any_path,
        upload_dandiset_metadata=upload_dandiset_metadata,
        devel_debug=devel_debug,
        jobs=jobs,
        jobs_per_file=jobs_per_file,
        sync=sync,
    )



from datetime import datetime, timezone
import re
from typing import Any, Callable, NoReturn

import numpy as np
from pynwb import NWBHDF5IO, NWBFile, TimeSeries

from ..pynwb_utils import _sanitize_nwb_version, nwb_has_external_links


def test_pynwb_io(simple1_nwb: str) -> None:
    # To verify that our dependencies spec is sufficient to avoid
    # stepping into known pynwb/hdmf issues
    with NWBHDF5IO(str(simple1_nwb), "r", load_namespaces=True) as reader:
        nwbfile = reader.read()
    assert repr(nwbfile)
    assert str(nwbfile)


def test_sanitize_nwb_version() -> None:
    def _nocall(*args: Any) -> NoReturn:
        raise AssertionError(f"Should have not been called. Was called with {args}")

    def assert_regex(regex: str) -> Callable[[str], None]:
        def search(v: str) -> None:
            assert re.search(regex, v)

        return search

    assert _sanitize_nwb_version("1.0.0", log=_nocall) == "1.0.0"
    assert _sanitize_nwb_version("NWB-1.0.0", log=_nocall) == "1.0.0"
    assert _sanitize_nwb_version("NWB-2.0.0", log=_nocall) == "2.0.0"
    assert (
        _sanitize_nwb_version(
            "NWB-2.1.0",
            log=assert_regex("^nwb_version 'NWB-2.1.0' starts with NWB- prefix,"),
        )
        == "2.1.0"
    )
    assert (
        _sanitize_nwb_version(
            "NWB-2.1.0",
            filename="/bu",
            log=assert_regex(
                "^File /bu: nwb_version 'NWB-2.1.0' starts with NWB- prefix,"
            ),
        )
        == "2.1.0"
    )


def test_nwb_has_external_links(tmp_path):
    # Create the base data
    start_time = datetime(2017, 4, 3, 11, tzinfo=timezone.utc)
    create_date = datetime(2017, 4, 15, 12, tzinfo=timezone.utc)
    data = np.arange(1000).reshape((100, 10))
    timestamps = np.arange(100)
    filename1 = tmp_path / "external1_example.nwb"
    filename4 = tmp_path / "external_linkdataset_example.nwb"

    # Create the first file
    nwbfile1 = NWBFile(
        session_description="demonstrate external files",
        identifier="NWBE1",
        session_start_time=start_time,
        file_create_date=create_date,
    )
    test_ts1 = TimeSeries(
        name="test_timeseries1", data=data, unit="SIunit", timestamps=timestamps
    )
    nwbfile1.add_acquisition(test_ts1)
    # Write the first file
    with NWBHDF5IO(filename1, "w") as io:
        io.write(nwbfile1)

    nwbfile4 = NWBFile(
        session_description="demonstrate external files",
        identifier="NWBE4",
        session_start_time=start_time,
        file_create_date=create_date,
    )

    # Get the first timeseries
    with NWBHDF5IO(filename1, "r") as io1:
        nwbfile1 = io1.read()
        timeseries_1_data = nwbfile1.get_acquisition("test_timeseries1").data

        # Create a new timeseries that links to our data
        test_ts4 = TimeSeries(
            name="test_timeseries4",
            data=timeseries_1_data,  # <-------
            unit="SIunit",
            timestamps=timestamps,
        )
        nwbfile4.add_acquisition(test_ts4)

        with NWBHDF5IO(filename4, "w") as io4:
            io4.write(nwbfile4, link_data=True)

    assert not nwb_has_external_links(filename1)
    assert nwb_has_external_links(filename4)



import json
import os
import os.path as op
from pathlib import Path
import re
from shutil import rmtree
from typing import Callable, List, Tuple

import numpy as np
import pytest
from pytest_mock import MockerFixture
import responses
import zarr

from .fixtures import SampleDandiset
from .skip import mark
from .test_helpers import assert_dirtrees_eq
from ..consts import DRAFT, dandiset_metadata_file
from ..dandiarchive import DandisetURL
from ..download import ProgressCombiner, download, download_generator
from ..utils import list_paths


# both urls point to 000027 (lean test dataset), and both draft and "released"
# version have only a single file ATM
@mark.skipif_no_network
@pytest.mark.parametrize(
    "url",
    [  # Should go through API
        "https://dandiarchive.org/dandiset/000027/0.210831.2033",
        # Drafts do not go through API ATM, but that should not be visible to user
        "https://dandiarchive.org/dandiset/000027/draft",
    ],
)
def test_download_000027(url: str, tmp_path: Path) -> None:
    ret = download(url, tmp_path)  # type: ignore[func-returns-value]
    assert not ret  # we return nothing ATM, might want to "generate"
    dsdir = tmp_path / "000027"
    assert list_paths(dsdir, dirs=True) == [
        dsdir / "dandiset.yaml",
        dsdir / "sub-RAT123",
        dsdir / "sub-RAT123" / "sub-RAT123.nwb",
    ]
    # and checksum should be correct as well
    from ..support.digests import Digester

    assert (
        Digester(["md5"])(dsdir / "sub-RAT123" / "sub-RAT123.nwb")["md5"]
        == "33318fd510094e4304868b4a481d4a5a"
    )
    # redownload - since already exist there should be an exception
    with pytest.raises(FileExistsError):
        download(url, tmp_path)

    # TODO: somehow get that status report about what was downloaded and what not
    download(url, tmp_path, existing="skip")  # TODO: check that skipped
    download(url, tmp_path, existing="overwrite")  # TODO: check that redownloaded
    download(url, tmp_path, existing="refresh")  # TODO: check that skipped (the same)


@mark.skipif_no_network
@pytest.mark.parametrize(
    "url",
    [  # Should go through API
        "https://dandiarchive.org/dandiset/000027/0.210831.2033",
        # Drafts do not go through API ATM, but that should not be visible to user
        "https://dandiarchive.org/dandiset/000027/draft",
    ],
)
def test_download_000027_metadata_only(url: str, tmp_path: Path) -> None:
    ret = download(url, tmp_path, get_assets=False)  # type: ignore[func-returns-value]
    assert not ret  # we return nothing ATM, might want to "generate"
    dsdir = tmp_path / "000027"
    assert list_paths(dsdir, dirs=True) == [dsdir / "dandiset.yaml"]


@mark.skipif_no_network
@pytest.mark.parametrize(
    "url",
    [  # Should go through API
        "https://dandiarchive.org/dandiset/000027/0.210831.2033",
        # Drafts do not go through API ATM, but that should not be visible to user
        "https://dandiarchive.org/dandiset/000027/draft",
    ],
)
def test_download_000027_assets_only(url: str, tmp_path: Path) -> None:
    ret = download(url, tmp_path, get_metadata=False)  # type: ignore[func-returns-value]
    assert not ret  # we return nothing ATM, might want to "generate"
    dsdir = tmp_path / "000027"
    assert list_paths(dsdir, dirs=True) == [
        dsdir / "sub-RAT123",
        dsdir / "sub-RAT123" / "sub-RAT123.nwb",
    ]


@mark.skipif_no_network
@pytest.mark.parametrize("resizer", [lambda sz: 0, lambda sz: sz // 2, lambda sz: sz])
@pytest.mark.parametrize("version", ["0.210831.2033", DRAFT])
def test_download_000027_resume(
    tmp_path: Path, resizer: Callable[[int], int], version: str
) -> None:
    from ..support.digests import Digester

    url = f"https://dandiarchive.org/dandiset/000027/{version}"
    digester = Digester()
    download(url, tmp_path, get_metadata=False)
    dsdir = tmp_path / "000027"
    nwb = dsdir / "sub-RAT123" / "sub-RAT123.nwb"
    digests = digester(str(nwb))
    dldir = nwb.with_name(nwb.name + ".dandidownload")
    dldir.mkdir()
    dlfile = dldir / "file"
    nwb.rename(dlfile)
    size = dlfile.stat().st_size
    os.truncate(dlfile, resizer(size))
    with (dldir / "checksum").open("w") as fp:
        json.dump(digests, fp)
    download(url, tmp_path, get_metadata=False)
    contents = [
        op.relpath(op.join(dirpath, entry), dsdir)
        for (dirpath, dirnames, filenames) in os.walk(dsdir)
        for entry in dirnames + filenames
    ]
    assert sorted(contents) == ["sub-RAT123", op.join("sub-RAT123", "sub-RAT123.nwb")]
    assert nwb.stat().st_size == size
    assert digester(str(nwb)) == digests


def test_download_newest_version(text_dandiset: SampleDandiset, tmp_path: Path) -> None:
    dandiset = text_dandiset.dandiset
    dandiset_id = text_dandiset.dandiset_id
    download(dandiset.api_url, tmp_path)
    assert (tmp_path / dandiset_id / "file.txt").read_text() == "This is test text.\n"
    dandiset.wait_until_valid()
    dandiset.publish()
    (text_dandiset.dspath / "file.txt").write_text("This is different text.\n")
    text_dandiset.upload()
    rmtree(tmp_path / dandiset_id)
    download(dandiset.api_url, tmp_path)
    assert (tmp_path / dandiset_id / "file.txt").read_text() == "This is test text.\n"


def test_download_folder(text_dandiset: SampleDandiset, tmp_path: Path) -> None:
    dandiset_id = text_dandiset.dandiset_id
    download(
        f"dandi://{text_dandiset.api.instance_id}/{dandiset_id}/subdir2/", tmp_path
    )
    assert list_paths(tmp_path, dirs=True) == [
        tmp_path / "subdir2",
        tmp_path / "subdir2" / "banana.txt",
        tmp_path / "subdir2" / "coconut.txt",
    ]
    assert (tmp_path / "subdir2" / "banana.txt").read_text() == "Banana\n"
    assert (tmp_path / "subdir2" / "coconut.txt").read_text() == "Coconut\n"


def test_download_item(text_dandiset: SampleDandiset, tmp_path: Path) -> None:
    dandiset_id = text_dandiset.dandiset_id
    download(
        f"dandi://{text_dandiset.api.instance_id}/{dandiset_id}/subdir2/coconut.txt",
        tmp_path,
    )
    assert list_paths(tmp_path, dirs=True) == [tmp_path / "coconut.txt"]
    assert (tmp_path / "coconut.txt").read_text() == "Coconut\n"


def test_download_asset_id(text_dandiset: SampleDandiset, tmp_path: Path) -> None:
    asset = text_dandiset.dandiset.get_asset_by_path("subdir2/coconut.txt")
    download(asset.download_url, tmp_path)
    assert list_paths(tmp_path, dirs=True) == [tmp_path / "coconut.txt"]
    assert (tmp_path / "coconut.txt").read_text() == "Coconut\n"


def test_download_asset_id_only(text_dandiset: SampleDandiset, tmp_path: Path) -> None:
    asset = text_dandiset.dandiset.get_asset_by_path("subdir2/coconut.txt")
    download(asset.base_download_url, tmp_path)
    assert list_paths(tmp_path, dirs=True) == [tmp_path / "coconut.txt"]
    assert (tmp_path / "coconut.txt").read_text() == "Coconut\n"


@pytest.mark.parametrize("confirm", [True, False])
def test_download_sync(
    confirm: bool, mocker: MockerFixture, text_dandiset: SampleDandiset, tmp_path: Path
) -> None:
    text_dandiset.dandiset.get_asset_by_path("file.txt").delete()
    dspath = tmp_path / text_dandiset.dandiset_id
    os.rename(text_dandiset.dspath, dspath)
    confirm_mock = mocker.patch(
        "dandi.download.abbrev_prompt", return_value="yes" if confirm else "no"
    )
    download(
        f"dandi://{text_dandiset.api.instance_id}/{text_dandiset.dandiset_id}",
        tmp_path,
        existing="overwrite",
        sync=True,
    )
    confirm_mock.assert_called_with("Delete 1 local asset?", "yes", "no", "list")
    if confirm:
        assert not (dspath / "file.txt").exists()
    else:
        assert (dspath / "file.txt").exists()


def test_download_sync_folder(
    mocker: MockerFixture, text_dandiset: SampleDandiset
) -> None:
    text_dandiset.dandiset.get_asset_by_path("file.txt").delete()
    text_dandiset.dandiset.get_asset_by_path("subdir2/banana.txt").delete()
    confirm_mock = mocker.patch("dandi.download.abbrev_prompt", return_value="yes")
    download(
        f"dandi://{text_dandiset.api.instance_id}/{text_dandiset.dandiset_id}/subdir2/",
        text_dandiset.dspath,
        existing="overwrite",
        sync=True,
    )
    confirm_mock.assert_called_with("Delete 1 local asset?", "yes", "no", "list")
    assert (text_dandiset.dspath / "file.txt").exists()
    assert not (text_dandiset.dspath / "subdir2" / "banana.txt").exists()


def test_download_sync_list(
    capsys: pytest.CaptureFixture[str],
    mocker: MockerFixture,
    text_dandiset: SampleDandiset,
    tmp_path: Path,
) -> None:
    text_dandiset.dandiset.get_asset_by_path("file.txt").delete()
    dspath = tmp_path / text_dandiset.dandiset_id
    os.rename(text_dandiset.dspath, dspath)
    input_mock = mocker.patch("dandi.utils.input", side_effect=["list", "yes"])
    download(
        f"dandi://{text_dandiset.api.instance_id}/{text_dandiset.dandiset_id}",
        tmp_path,
        existing="overwrite",
        sync=True,
    )
    assert not (dspath / "file.txt").exists()
    assert input_mock.call_args_list == [
        mocker.call("Delete 1 local asset? ([y]es/[n]o/[l]ist): "),
        mocker.call("Delete 1 local asset? ([y]es/[n]o/[l]ist): "),
    ]
    assert capsys.readouterr().out.splitlines()[-1] == str(dspath / "file.txt")


def test_download_sync_zarr(
    mocker: MockerFixture, zarr_dandiset: SampleDandiset, tmp_path: Path
) -> None:
    zarr_dandiset.dandiset.get_asset_by_path("sample.zarr").delete()
    dspath = tmp_path / zarr_dandiset.dandiset_id
    os.rename(zarr_dandiset.dspath, dspath)
    confirm_mock = mocker.patch("dandi.download.abbrev_prompt", return_value="yes")
    download(
        zarr_dandiset.dandiset.version_api_url,
        tmp_path,
        existing="overwrite",
        sync=True,
    )
    confirm_mock.assert_called_with("Delete 1 local asset?", "yes", "no", "list")
    assert not (dspath / "sample.zarr").exists()


@responses.activate
def test_download_no_blobDateModified(
    text_dandiset: SampleDandiset, tmp_path: Path
) -> None:
    # Regression test for #806
    responses.add_passthru(re.compile("^http"))
    dandiset = text_dandiset.dandiset
    asset = dandiset.get_asset_by_path("file.txt")
    metadata = asset.get_raw_metadata()
    del metadata["blobDateModified"]
    responses.add(responses.GET, asset.api_url, json=metadata)
    download(dandiset.api_url, tmp_path)


@responses.activate
def test_download_metadata404(text_dandiset: SampleDandiset, tmp_path: Path) -> None:
    responses.add_passthru(re.compile("^http"))
    asset = text_dandiset.dandiset.get_asset_by_path("subdir1/apple.txt")
    responses.add(responses.GET, asset.api_url, status=404)
    statuses = list(
        download_generator(
            DandisetURL(
                api_url=text_dandiset.client.api_url,
                dandiset_id=text_dandiset.dandiset.identifier,
                version_id=text_dandiset.dandiset.version_id,
            ),
            tmp_path,
        )
    )
    errors = [s for s in statuses if s.get("status") == "error"]
    assert errors == [
        {
            "path": "subdir1/apple.txt",
            "status": "error",
            "message": f"No such asset: {asset}",
        }
    ]
    assert list_paths(tmp_path, dirs=True) == [
        tmp_path / dandiset_metadata_file,
        tmp_path / "file.txt",
        tmp_path / "subdir2",
        tmp_path / "subdir2" / "banana.txt",
        tmp_path / "subdir2" / "coconut.txt",
    ]


def test_download_zarr(tmp_path: Path, zarr_dandiset: SampleDandiset) -> None:
    download(zarr_dandiset.dandiset.version_api_url, tmp_path)
    assert_dirtrees_eq(
        zarr_dandiset.dspath / "sample.zarr",
        tmp_path / zarr_dandiset.dandiset_id / "sample.zarr",
    )


def test_download_different_zarr(tmp_path: Path, zarr_dandiset: SampleDandiset) -> None:
    dd = tmp_path / zarr_dandiset.dandiset_id
    dd.mkdir()
    zarr.save(dd / "sample.zarr", np.eye(5))
    download(
        zarr_dandiset.dandiset.version_api_url, tmp_path, existing="overwrite-different"
    )
    assert_dirtrees_eq(
        zarr_dandiset.dspath / "sample.zarr",
        tmp_path / zarr_dandiset.dandiset_id / "sample.zarr",
    )


def test_download_different_zarr_delete_dir(
    new_dandiset: SampleDandiset, tmp_path: Path
) -> None:
    d = new_dandiset.dandiset
    dspath = new_dandiset.dspath
    zarr.save(dspath / "sample.zarr", np.eye(5))
    assert not any(p.is_dir() for p in (dspath / "sample.zarr").iterdir())
    new_dandiset.upload()
    dd = tmp_path / d.identifier
    dd.mkdir(parents=True, exist_ok=True)
    zarr.save(dd / "sample.zarr", np.arange(1000), np.arange(1000, 0, -1))
    assert any(p.is_dir() for p in (dd / "sample.zarr").iterdir())
    download(d.version_api_url, tmp_path, existing="overwrite-different")
    assert_dirtrees_eq(dspath / "sample.zarr", dd / "sample.zarr")


def test_download_zarr_to_nonzarr_path(
    tmp_path: Path, zarr_dandiset: SampleDandiset
) -> None:
    dd = tmp_path / zarr_dandiset.dandiset_id
    dd.mkdir()
    (dd / "sample.zarr").write_text("This is not a Zarr.\n")
    download(
        zarr_dandiset.dandiset.version_api_url, tmp_path, existing="overwrite-different"
    )
    assert_dirtrees_eq(
        zarr_dandiset.dspath / "sample.zarr",
        tmp_path / zarr_dandiset.dandiset_id / "sample.zarr",
    )


def test_download_nonzarr_to_zarr_path(
    new_dandiset: SampleDandiset, tmp_path: Path
) -> None:
    d = new_dandiset.dandiset
    (new_dandiset.dspath / "sample.zarr").write_text("This is not a Zarr.\n")
    new_dandiset.upload(allow_any_path=True)
    dd = tmp_path / d.identifier
    dd.mkdir(parents=True, exist_ok=True)
    zarr.save(dd / "sample.zarr", np.arange(1000), np.arange(1000, 0, -1))
    download(d.version_api_url, tmp_path, existing="overwrite-different")
    assert (dd / "sample.zarr").is_file()
    assert (dd / "sample.zarr").read_text() == "This is not a Zarr.\n"


def test_download_zarr_asset_id_only(
    zarr_dandiset: SampleDandiset, tmp_path: Path
) -> None:
    asset = zarr_dandiset.dandiset.get_asset_by_path("sample.zarr")
    download(asset.base_download_url, tmp_path)
    assert list(tmp_path.iterdir()) == [tmp_path / "sample.zarr"]
    assert_dirtrees_eq(zarr_dandiset.dspath / "sample.zarr", tmp_path / "sample.zarr")


def test_download_zarr_subdir_has_only_subdirs(
    tmp_path: Path, new_dandiset: SampleDandiset
) -> None:
    zf = new_dandiset.dspath / "sample.zarr"
    zf.mkdir()
    (zf / "dirs").mkdir()
    (zf / "dirs" / "apple").mkdir()
    (zf / "dirs" / "apple" / "file.txt").write_text("Apple\n")
    (zf / "dirs" / "banana").mkdir()
    (zf / "dirs" / "banana" / "file.txt").write_text("Banana\n")
    (zf / "dirs" / "coconut").mkdir()
    (zf / "dirs" / "coconut" / "file.txt").write_text("Coconut\n")
    new_dandiset.upload(validation="skip")
    download(new_dandiset.dandiset.version_api_url, tmp_path)
    assert_dirtrees_eq(zf, tmp_path / new_dandiset.dandiset_id / "sample.zarr")


@pytest.mark.parametrize(
    "file_qty,inputs,expected",
    [
        (
            1,
            [
                ("lonely.txt", {"size": 42}),
                ("lonely.txt", {"status": "downloading"}),
                ("lonely.txt", {"done": 0, "done%": 0.0}),
                ("lonely.txt", {"done": 20, "done%": 20 / 42 * 100}),
                ("lonely.txt", {"done": 40, "done%": 40 / 42 * 100}),
                ("lonely.txt", {"done": 42, "done%": 100.0}),
                ("lonely.txt", {"checksum": "ok"}),
                ("lonely.txt", {"status": "setting mtime"}),
                ("lonely.txt", {"status": "done"}),
            ],
            [
                {"size": 69105},
                {"status": "downloading"},
                {"done": 0, "done%": 0.0},
                {"done": 20, "done%": 20 / 42 * 100},
                {"done": 40, "done%": 40 / 42 * 100},
                {"done": 42, "done%": 100.0},
                {"status": "done", "message": "1 done"},
            ],
        ),
        (
            2,
            [
                ("apple.txt", {"size": 42}),
                ("banana.txt", {"size": 127}),
                ("apple.txt", {"status": "downloading"}),
                ("banana.txt", {"status": "downloading"}),
                ("apple.txt", {"done": 0, "done%": 0.0}),
                ("banana.txt", {"done": 0, "done%": 0.0}),
                ("apple.txt", {"done": 20, "done%": 20 / 42 * 100}),
                ("banana.txt", {"done": 40, "done%": 40 / 127 * 100}),
                ("apple.txt", {"done": 40, "done%": 40 / 42 * 100}),
                ("banana.txt", {"done": 80, "done%": 80 / 127 * 100}),
                ("apple.txt", {"done": 42, "done%": 100.0}),
                ("banana.txt", {"done": 120, "done%": 120 / 127 * 100}),
                ("apple.txt", {"checksum": "ok"}),
                ("banana.txt", {"done": 127, "done%": 100.0}),
                ("apple.txt", {"status": "setting mtime"}),
                ("banana.txt", {"checksum": "ok"}),
                ("apple.txt", {"status": "done"}),
                ("banana.txt", {"status": "setting mtime"}),
                ("banana.txt", {"status": "done"}),
            ],
            [
                {"size": 69105},
                {"status": "downloading"},
                {"done": 0, "done%": 0.0},
                {"done": 0, "done%": 0.0},
                {"done": 20, "done%": 20 / 169 * 100},
                {"done": 60, "done%": 60 / 169 * 100},
                {"done": 80, "done%": 80 / 169 * 100},
                {"done": 120, "done%": 120 / 169 * 100},
                {"done": 122, "done%": 122 / 169 * 100},
                {"done": 162, "done%": 162 / 169 * 100},
                {"done": 169, "done%": 100.0},
                {"message": "1 done"},
                {"status": "done", "message": "2 done"},
            ],
        ),
        (
            2,
            [
                ("apple.txt", {"size": 42}),
                ("apple.txt", {"status": "downloading"}),
                ("apple.txt", {"done": 0, "done%": 0.0}),
                ("apple.txt", {"done": 20, "done%": 20 / 42 * 100}),
                ("banana.txt", {"size": 127}),
                ("apple.txt", {"done": 40, "done%": 40 / 42 * 100}),
                ("banana.txt", {"status": "downloading"}),
                ("apple.txt", {"done": 42, "done%": 100.0}),
                ("banana.txt", {"done": 0, "done%": 0.0}),
                ("apple.txt", {"checksum": "ok"}),
                ("banana.txt", {"done": 40, "done%": 40 / 127 * 100}),
                ("apple.txt", {"status": "setting mtime"}),
                ("banana.txt", {"done": 80, "done%": 80 / 127 * 100}),
                ("apple.txt", {"status": "done"}),
                ("banana.txt", {"done": 120, "done%": 120 / 127 * 100}),
                ("banana.txt", {"done": 127, "done%": 100.0}),
                ("banana.txt", {"checksum": "ok"}),
                ("banana.txt", {"status": "setting mtime"}),
                ("banana.txt", {"status": "done"}),
            ],
            [
                {"size": 69105},
                {"status": "downloading"},
                {"done": 0, "done%": 0.0},
                {"done": 20, "done%": 20 / 42 * 100},
                {"done": 20, "done%": 20 / 169 * 100},
                {"done": 40, "done%": 40 / 169 * 100},
                {"done": 42, "done%": 42 / 169 * 100},
                {"done": 42, "done%": 42 / 169 * 100},
                {"done": 82, "done%": 82 / 169 * 100},
                {"done": 122, "done%": 122 / 169 * 100},
                {"message": "1 done"},
                {"done": 162, "done%": 162 / 169 * 100},
                {"done": 169, "done%": 169 / 169 * 100},
                {"status": "done", "message": "2 done"},
            ],
        ),
        (
            2,
            [
                ("apple.txt", {"size": 42}),
                ("apple.txt", {"status": "downloading"}),
                ("apple.txt", {"done": 0, "done%": 0.0}),
                ("apple.txt", {"done": 20, "done%": 20 / 42 * 100}),
                ("apple.txt", {"done": 40, "done%": 40 / 42 * 100}),
                ("apple.txt", {"done": 42, "done%": 100.0}),
                ("apple.txt", {"checksum": "ok"}),
                ("apple.txt", {"status": "setting mtime"}),
                ("apple.txt", {"status": "done"}),
                ("banana.txt", {"size": 127}),
                ("banana.txt", {"status": "downloading"}),
                ("banana.txt", {"done": 0, "done%": 0.0}),
                ("banana.txt", {"done": 40, "done%": 40 / 127 * 100}),
                ("banana.txt", {"done": 80, "done%": 80 / 127 * 100}),
                ("banana.txt", {"done": 120, "done%": 120 / 127 * 100}),
                ("banana.txt", {"done": 127, "done%": 100.0}),
                ("banana.txt", {"checksum": "ok"}),
                ("banana.txt", {"status": "setting mtime"}),
                ("banana.txt", {"status": "done"}),
            ],
            [
                {"size": 69105},
                {"status": "downloading"},
                {"done": 0, "done%": 0.0},
                {"done": 20, "done%": 20 / 42 * 100},
                {"done": 40, "done%": 40 / 42 * 100},
                {"done": 42, "done%": 42 / 42 * 100},
                {"message": "1 done"},
                {"done": 42, "done%": 42 / 169 * 100},
                {"done": 82, "done%": 82 / 169 * 100},
                {"done": 122, "done%": 122 / 169 * 100},
                {"done": 162, "done%": 162 / 169 * 100},
                {"done": 169, "done%": 100.0},
                {"status": "done", "message": "2 done"},
            ],
        ),
        (
            2,
            [
                ("apple.txt", {"size": 42}),
                ("banana.txt", {"size": 127}),
                ("apple.txt", {"status": "downloading"}),
                ("banana.txt", {"status": "downloading"}),
                ("apple.txt", {"done": 0, "done%": 0.0}),
                ("banana.txt", {"done": 0, "done%": 0.0}),
                ("apple.txt", {"done": 20, "done%": 20 / 42 * 100}),
                ("banana.txt", {"done": 40, "done%": 40 / 127 * 100}),
                ("apple.txt", {"done": 40, "done%": 40 / 42 * 100}),
                ("banana.txt", {"status": "error", "message": "Internet broke"}),
                ("apple.txt", {"done": 42, "done%": 100.0}),
                ("apple.txt", {"checksum": "ok"}),
                ("apple.txt", {"status": "setting mtime"}),
                ("apple.txt", {"status": "done"}),
            ],
            [
                {"size": 69105},
                {"status": "downloading"},
                {"done": 0, "done%": 0.0},
                {"done": 0, "done%": 0.0},
                {"done": 20, "done%": 20 / 169 * 100},
                {"done": 60, "done%": 60 / 169 * 100},
                {"done": 80, "done%": 80 / 169 * 100},
                {"message": "1 errored"},
                {"done": 40, "done%": 40 / 42 * 100},
                {"done": 42, "done%": 100.0},
                {"status": "error", "message": "1 done, 1 errored"},
            ],
        ),
        (
            1,
            [("lonely.txt", {"status": "skipped", "message": "already exists"})],
            [{"status": "skipped", "message": "1 skipped"}],
        ),
        (
            2,
            [
                ("apple.txt", {"size": 42}),
                ("banana.txt", {"status": "skipped", "message": "already exists"}),
                ("apple.txt", {"status": "downloading"}),
                ("apple.txt", {"done": 0, "done%": 0.0}),
                ("apple.txt", {"done": 20, "done%": 20 / 42 * 100}),
                ("apple.txt", {"done": 40, "done%": 40 / 42 * 100}),
                ("apple.txt", {"done": 42, "done%": 100.0}),
                ("apple.txt", {"checksum": "ok"}),
                ("apple.txt", {"status": "setting mtime"}),
                ("apple.txt", {"status": "done"}),
            ],
            [
                {"size": 69105},
                {"message": "1 skipped"},
                {"status": "downloading"},
                {"done": 0, "done%": 0.0},
                {"done": 20, "done%": 20 / 42 * 100},
                {"done": 40, "done%": 40 / 42 * 100},
                {"done": 42, "done%": 100.0},
                {"status": "done", "message": "1 done, 1 skipped"},
            ],
        ),
        (
            2,
            [
                ("apple.txt", {"size": 42}),
                ("banana.txt", {"size": 127}),
                ("apple.txt", {"status": "downloading"}),
                ("banana.txt", {"status": "downloading"}),
                ("apple.txt", {"done": 0, "done%": 0.0}),
                ("banana.txt", {"done": 0, "done%": 0.0}),
                ("apple.txt", {"done": 20, "done%": 20 / 42 * 100}),
                ("banana.txt", {"done": 40, "done%": 40 / 127 * 100}),
                ("apple.txt", {"done": 40, "done%": 40 / 42 * 100}),
                ("banana.txt", {"done": 80, "done%": 80 / 127 * 100}),
                ("apple.txt", {"done": 42, "done%": 100.0}),
                ("banana.txt", {"done": 120, "done%": 120 / 127 * 100}),
                ("apple.txt", {"checksum": "ok"}),
                ("banana.txt", {"done": 127, "done%": 100.0}),
                ("apple.txt", {"status": "setting mtime"}),
                (
                    "banana.txt",
                    {
                        "checksum": "differs",
                        "status": "error",
                        "message": "Checksum differs",
                    },
                ),
                ("apple.txt", {"status": "done"}),
            ],
            [
                {"size": 69105},
                {"status": "downloading"},
                {"done": 0, "done%": 0.0},
                {"done": 0, "done%": 0.0},
                {"done": 20, "done%": 20 / 169 * 100},
                {"done": 60, "done%": 60 / 169 * 100},
                {"done": 80, "done%": 80 / 169 * 100},
                {"done": 120, "done%": 120 / 169 * 100},
                {"done": 122, "done%": 122 / 169 * 100},
                {"done": 162, "done%": 162 / 169 * 100},
                {"done": 169, "done%": 100.0},
                {"message": "1 errored"},
                {"status": "error", "message": "1 done, 1 errored"},
            ],
        ),
        (
            3,
            [
                ("apple.txt", {"size": 42}),
                ("banana.txt", {"size": 127}),
                ("apple.txt", {"status": "downloading"}),
                ("banana.txt", {"status": "downloading"}),
                ("coconut", {"status": "skipped", "message": "already exists"}),
                ("apple.txt", {"done": 0, "done%": 0.0}),
                ("banana.txt", {"done": 0, "done%": 0.0}),
                ("apple.txt", {"done": 20, "done%": 20 / 42 * 100}),
                ("banana.txt", {"done": 40, "done%": 40 / 127 * 100}),
                ("apple.txt", {"done": 40, "done%": 40 / 42 * 100}),
                ("banana.txt", {"done": 80, "done%": 80 / 127 * 100}),
                ("apple.txt", {"done": 42, "done%": 100.0}),
                (
                    "apple.txt",
                    {
                        "checksum": "differs",
                        "status": "error",
                        "message": "Checksum differs",
                    },
                ),
                ("banana.txt", {"done": 120, "done%": 120 / 127 * 100}),
                ("banana.txt", {"done": 127, "done%": 100.0}),
                ("banana.txt", {"checksum": "ok"}),
                ("banana.txt", {"status": "setting mtime"}),
                ("banana.txt", {"status": "done"}),
            ],
            [
                {"size": 69105},
                {"status": "downloading"},
                {"message": "1 skipped"},
                {"done": 0, "done%": 0.0},
                {"done": 0, "done%": 0.0},
                {"done": 20, "done%": 20 / 169 * 100},
                {"done": 60, "done%": 60 / 169 * 100},
                {"done": 80, "done%": 80 / 169 * 100},
                {"done": 120, "done%": 120 / 169 * 100},
                {"done": 122, "done%": 122 / 169 * 100},
                {"message": "1 errored, 1 skipped"},
                {"done": 162, "done%": 162 / 169 * 100},
                {"done": 169, "done%": 100.0},
                {"status": "error", "message": "1 done, 1 errored, 1 skipped"},
            ],
        ),
    ],
)
def test_progress_combiner(
    file_qty: int, inputs: List[Tuple[str, dict]], expected: List[dict]
) -> None:
    pc = ProgressCombiner(zarr_size=69105, file_qty=file_qty)
    outputs: List[dict] = []
    for path, status in inputs:
        outputs.extend(pc.feed(path, status))
    assert outputs == expected



from collections import defaultdict
from functools import reduce
import os.path
from pathlib import Path
import re
import sys
import time
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional, Set, Tuple, Union

import click

from . import lgr
from .consts import DRAFT, dandiset_identifier_regex, dandiset_metadata_file
from .dandiapi import RemoteAsset
from .exceptions import NotFoundError
from .files import (
    DandiFile,
    DandisetMetadataFile,
    LocalAsset,
    LocalDirectoryAsset,
    ZarrAsset,
    find_dandi_files,
)
from .misctypes import Digest
from .utils import ensure_datetime, get_instance, pluralize

if TYPE_CHECKING:
    if sys.version_info >= (3, 8):
        from typing import TypedDict
    else:
        from typing_extensions import TypedDict

    class Uploaded(TypedDict):
        size: int
        errors: List[str]


def upload(
    paths: Optional[List[Union[str, Path]]] = None,
    existing: str = "refresh",
    validation: str = "require",
    dandi_instance: str = "dandi",
    allow_any_path: bool = False,
    upload_dandiset_metadata: bool = False,
    devel_debug: bool = False,
    jobs: Optional[int] = None,
    jobs_per_file: Optional[int] = None,
    sync: bool = False,
) -> None:
    from .dandiapi import DandiAPIClient
    from .dandiset import APIDandiset, Dandiset

    if paths:
        paths = [Path(p).absolute() for p in paths]
        dandiset_ = Dandiset.find(os.path.commonpath(paths))
    else:
        dandiset_ = Dandiset.find(None)
    if not dandiset_:
        raise RuntimeError(
            f"Found no {dandiset_metadata_file} anywhere in common ancestor of"
            " paths.  Use 'dandi download' or 'organize' first."
        )

    instance = get_instance(dandi_instance)
    assert instance.api is not None
    api_url = instance.api

    # We need to use this as a context manager in order to ensure the session
    # gets properly closed.  Otherwise, pytest sometimes complains under
    # obscure conditions.
    with DandiAPIClient(api_url) as client:
        client.check_schema_version()
        client.dandi_authenticate()

        dandiset = APIDandiset(dandiset_.path)  # "cast" to a new API based dandiset

        ds_identifier = dandiset.identifier
        remote_dandiset = client.get_dandiset(ds_identifier, DRAFT)

        if not re.match(dandiset_identifier_regex, str(ds_identifier)):
            raise ValueError(
                f"Dandiset identifier {ds_identifier} does not follow expected "
                f"convention {dandiset_identifier_regex!r}."
            )

        from .pynwb_utils import ignore_benign_pynwb_warnings
        from .support.pyout import naturalsize
        from .utils import path_is_subpath

        ignore_benign_pynwb_warnings()  # so validate doesn't whine

        if not paths:
            paths = [dandiset.path]

        dandi_files = list(
            find_dandi_files(
                *paths,
                dandiset_path=dandiset.path,
                allow_all=allow_any_path,
                include_metadata=True,
            )
        )
        lgr.info(f"Found {len(dandi_files)} files to consider")

        # We will keep a shared set of "being processed" paths so
        # we could limit the number of them until
        #   https://github.com/pyout/pyout/issues/87
        # properly addressed
        process_paths: Set[str] = set()

        uploaded_paths: Dict[str, Uploaded] = defaultdict(
            lambda: {"size": 0, "errors": []}
        )

        # TODO: we might want to always yield a full record so no field is not
        # provided to pyout to cause it to halt
        def process_path(dfile: DandiFile) -> Iterator[dict]:
            """

            Parameters
            ----------
            dfile: DandiFile

            Yields
            ------
            dict
              Records for pyout
            """
            strpath = str(dfile.filepath)
            try:
                if not isinstance(dfile, LocalDirectoryAsset):
                    try:
                        yield {"size": dfile.size}
                    except FileNotFoundError:
                        yield skip_file("ERROR: File not found")
                        return
                    except Exception as exc:
                        # without limiting [:50] it might cause some pyout indigestion
                        yield skip_file("ERROR: %s" % str(exc)[:50])
                        return

                #
                # Validate first, so we do not bother server at all if not kosher
                #
                # TODO: enable back validation of dandiset.yaml
                if isinstance(dfile, LocalAsset) and validation != "skip":
                    yield {"status": "pre-validating"}
                    validation_errors = dfile.get_validation_errors()
                    yield {"errors": len(validation_errors)}
                    # TODO: split for dandi, pynwb errors
                    if validation_errors:
                        if validation == "require":
                            yield skip_file("failed validation")
                            return
                    else:
                        yield {"status": "validated"}
                else:
                    # yielding empty causes pyout to get stuck or crash
                    # https://github.com/pyout/pyout/issues/91
                    # yield {"errors": '',}
                    pass

                #
                # Special handling for dandiset.yaml
                # Yarik hates it but that is life for now. TODO
                #
                if isinstance(dfile, DandisetMetadataFile):
                    # TODO This is a temporary measure to avoid breaking web UI
                    # dandiset metadata schema assumptions.  All edits should happen
                    # online.
                    if upload_dandiset_metadata:
                        yield {"status": "updating metadata"}
                        assert dandiset.metadata is not None
                        remote_dandiset.set_raw_metadata(dandiset.metadata)
                        yield {"status": "updated metadata"}
                    else:
                        yield skip_file("should be edited online")
                    return
                assert isinstance(dfile, LocalAsset)

                #
                # Compute checksums
                #
                file_etag: Optional[Digest]
                if isinstance(dfile, ZarrAsset):
                    file_etag = None
                else:
                    yield {"status": "digesting"}
                    try:
                        file_etag = dfile.get_digest()
                    except Exception as exc:
                        yield skip_file("failed to compute digest: %s" % str(exc))
                        return

                try:
                    extant = remote_dandiset.get_asset_by_path(dfile.path)
                except NotFoundError:
                    extant = None
                else:
                    assert extant is not None
                    replace, out = check_replace_asset(
                        local_asset=dfile,
                        remote_asset=extant,
                        existing=existing,
                        local_etag=file_etag,
                    )
                    yield out
                    if not replace:
                        return

                #
                # Extract metadata - delayed since takes time, but is done before
                # actual upload, so we could skip if this fails
                #
                # Extract metadata before actual upload and skip if fails
                # TODO: allow for for non-nwb files to skip this step
                # ad-hoc for dandiset.yaml for now
                yield {"status": "extracting metadata"}
                try:
                    metadata = dfile.get_metadata(
                        digest=file_etag, ignore_errors=allow_any_path
                    ).json_dict()
                except Exception as e:
                    yield skip_file("failed to extract metadata: %s" % str(e))
                    return

                #
                # Upload file
                #
                yield {"status": "uploading"}
                validating = False
                for r in dfile.iter_upload(
                    remote_dandiset, metadata, jobs=jobs_per_file, replacing=extant
                ):
                    r.pop("asset", None)  # to keep pyout from choking
                    if r["status"] == "uploading":
                        uploaded_paths[strpath]["size"] = r.pop("current")
                        yield r
                    elif r["status"] == "post-validating":
                        # Only yield the first "post-validating" status
                        if not validating:
                            yield r
                            validating = True
                    else:
                        yield r
                yield {"status": "done"}

            except Exception as exc:
                if devel_debug:
                    raise
                lgr.exception("Error uploading %s:", strpath)
                # Custom formatting for some exceptions we know to extract
                # user-meaningful message
                message = str(exc)
                uploaded_paths[strpath]["errors"].append(message)
                yield {"status": "ERROR", "message": message}
            finally:
                process_paths.remove(strpath)

        # We will again use pyout to provide a neat table summarizing our progress
        # with upload etc
        from .support import pyout as pyouts

        # for the upload speeds we need to provide a custom  aggregate
        t0 = time.time()

        def upload_agg(*ignored: Any) -> str:
            dt = time.time() - t0
            # to help avoiding dict length changes during upload
            # might be not a proper solution
            # see https://github.com/dandi/dandi-cli/issues/502 for more info
            uploaded_recs = list(uploaded_paths.values())
            total = sum(v["size"] for v in uploaded_recs)
            if not total:
                return ""
            speed = total / dt if dt else 0
            return "%s/s" % naturalsize(speed)

        pyout_style = pyouts.get_style(hide_if_missing=False)
        pyout_style["upload"]["aggregate"] = upload_agg

        rec_fields = ["path", "size", "errors", "upload", "status", "message"]
        out = pyouts.LogSafeTabular(
            style=pyout_style, columns=rec_fields, max_workers=jobs or 5
        )

        with out:
            for dfile in dandi_files:
                while len(process_paths) >= 10:
                    lgr.log(2, "Sleep waiting for some paths to finish processing")
                    time.sleep(0.5)

                process_paths.add(str(dfile.filepath))

                rec: Dict[Any, Any]
                if isinstance(dfile, DandisetMetadataFile):
                    rec = {"path": dandiset_metadata_file}
                else:
                    assert isinstance(dfile, LocalAsset)
                    rec = {"path": dfile.path}

                try:
                    if devel_debug:
                        # DEBUG: do serially
                        for v in process_path(dfile):
                            print(str(v), flush=True)
                    else:
                        rec[tuple(rec_fields[1:])] = process_path(dfile)
                except ValueError as exc:
                    rec.update(skip_file(exc))
                out(rec)

        if sync:
            relpaths: List[str] = []
            for p in paths:
                rp = os.path.relpath(p, dandiset.path)
                relpaths.append("" if rp == "." else rp)
            path_prefix = reduce(os.path.commonprefix, relpaths)  # type: ignore[arg-type]
            to_delete = []
            for asset in remote_dandiset.get_assets_with_path_prefix(path_prefix):
                if (
                    any(p == "" or path_is_subpath(asset.path, p) for p in relpaths)
                    and not Path(dandiset.path, asset.path).exists()
                ):
                    to_delete.append(asset)
            if to_delete and click.confirm(
                f"Delete {pluralize(len(to_delete), 'asset')} on server?"
            ):
                for asset in to_delete:
                    asset.delete()


def check_replace_asset(
    local_asset: LocalAsset,
    remote_asset: RemoteAsset,
    existing: str,
    local_etag: Optional[Digest],
) -> Tuple[bool, Dict[str, str]]:
    # Returns a (replace asset, message to yield) tuple
    if isinstance(local_asset, ZarrAsset):
        return (True, {"message": "exists - reuploading"})
    assert local_etag is not None
    metadata = remote_asset.get_raw_metadata()
    local_mtime = local_asset.modified
    remote_mtime_str = metadata.get("blobDateModified")
    # TODO: Should this error if the digest is missing?
    remote_etag = metadata.get("digest", {}).get(local_etag.algorithm.value)
    if remote_mtime_str is not None:
        remote_mtime = ensure_datetime(remote_mtime_str)
        remote_file_status = (
            "same"
            if remote_etag == local_etag.value and remote_mtime == local_mtime
            else (
                "newer"
                if remote_mtime > local_mtime
                else ("older" if remote_mtime < local_mtime else "diff")
            )
        )
    else:
        remote_mtime = None
        remote_file_status = "no mtime"
    exists_msg = f"exists ({remote_file_status})"
    if existing == "error":
        # as promised -- not gentle at all!
        raise FileExistsError(exists_msg)
    if existing == "skip":
        return (False, skip_file(exists_msg))
    # Logic below only for overwrite and reupload
    if existing == "overwrite":
        if remote_etag == local_etag.value:
            return (False, skip_file(exists_msg))
    elif existing == "refresh":
        if remote_etag == local_etag.value:
            return (False, skip_file("file exists"))
        elif remote_mtime is not None and remote_mtime >= local_mtime:
            return (False, skip_file(exists_msg))
    elif existing == "force":
        pass
    else:
        raise ValueError(f"invalid value for 'existing': {existing!r}")
    return (True, {"message": f"{exists_msg} - reuploading"})


def skip_file(msg: Any) -> Dict[str, str]:
    return {"status": "skipped", "message": str(msg)}
